# -*- coding: utf-8 -*-
import keyword
import os
import re
from pathlib import PurePosixPath, Path
from typing import Union, Iterable, List, Dict, Set

import yaml
from dotty_dict import Dotty
from ddb.utils.simpleeval import simple_eval

from ddb.feature import features
from ddb.feature.traefik import TraefikExtraServicesAction
from .binaries import DockerBinary
from .lib.compose.config.types import ServicePort
from ...action import Action
from ...action.action import EventBinding, InitializableAction
from ...binary import binaries, Binary
from ...cache import caches, register_project_cache
from ...config import config
from ...context import context
from ...event import bus, events
from ...utils.process import run
from ...utils.table_display import get_table_display


class EmitDockerComposeConfigAction(Action):
    """
    Emit docker:docker-compose-config event with docker compose configuration,
    and events from ddb.event.bus.emit.<event-name>=prop1=prop1_value;prop2=int(prop2_value) labels.
    To generate multiple events of same name in the same service, event-name can be suffixed with "[xxx]".
    """

    def __init__(self):
        super().__init__()
        self.current_yaml_output = None
        self.key_re = re.compile(r"^\s*ddb\.emit\.(.+?)(?:\[(.+?)\])?(?:\((.+?)\))?\s*$")
        self.eval_re = re.compile(r"^\s*eval\((.*)\)\s*$")
        self._cert_domains_cache = register_project_cache("docker.cert-domains")

    @property
    def event_bindings(self):
        return (
            # TODO: Add support for custom docker-compose -f option (custom filename and multiple files)
            EventBinding(events.file.found,
                         processor=lambda file: ((), {}) if file == "docker-compose.yml" else False),
            EventBinding(events.file.generated,
                         processor=lambda source, target: ((), {}) if target == "docker-compose.yml" else False)
        )

    @property
    def name(self) -> str:
        return "docker:emit-docker-compose-config"

    def execute(self):
        """
        Execute action
        """
        if not os.path.exists("docker-compose.yml"):
            return

        yaml_output = run("docker-compose", "config")

        parsed_config = yaml.load(yaml_output, yaml.SafeLoader)
        docker_compose_config = Dotty(parsed_config)

        if self.current_yaml_output == yaml_output:
            return
        self.current_yaml_output = yaml_output

        events.docker.docker_compose_config(docker_compose_config=docker_compose_config)

        services = docker_compose_config.get('services')
        if not services:
            return

        events.docker.docker_compose_before_events(docker_compose_config=docker_compose_config)

        cert_domains = []

        def on_available(domain: str, wildcard: bool, private_key: Union[bytes, str], certificate: Union[bytes, str]):
            """
            When a certificate is available.
            :param domain:
            :param wildcard:
            :param private_key:
            :param certificate:
            :return:
            """
            cert_domains.append(domain)

        off = bus.on("certs:available", on_available)
        try:
            self._parse_docker_compose(docker_compose_config, services)
        finally:
            off()

        self._update_cache_and_emit_certs_remove(cert_domains)

        events.docker.docker_compose_after_events(docker_compose_config=docker_compose_config)

    def _parse_docker_compose(self, docker_compose_config: dict, services):
        for service_name, service in services.items():
            labels = service.get('labels')
            if not labels:
                continue

            if not isinstance(labels, dict):
                labels = {label.split("=", 1) for label in labels}

            event_data = self._build_event_data(docker_compose_config, labels, service, service_name)

            for (event_name, event_parsed_values) in event_data.items():
                for _, (args, kwargs) in event_parsed_values.items():
                    bus.emit(event_name, *args, **kwargs)

    def _update_cache_and_emit_certs_remove(self, cert_domains: Iterable[str]):
        removed_domains = []
        for previous_cert_domain in self._cert_domains_cache.keys():
            if previous_cert_domain not in cert_domains:
                events.certs.remove(previous_cert_domain)
                removed_domains.append(previous_cert_domain)

        for removed_domain in removed_domains:
            self._cert_domains_cache.pop(removed_domain)

        for cert_domain in cert_domains:
            self._cert_domains_cache.set(cert_domain, None)

    def _build_event_data(self, docker_compose_config, labels, service, service_name):  # pylint:disable=too-many-locals
        parsed_values = {}
        for key, value in labels.items():
            match = self.key_re.match(key)
            if not match:
                continue

            event_name = match.group(1)
            event_id = match.group(2)
            property_name = match.group(3)

            if property_name and property_name in keyword.kwlist:
                property_name = property_name + '_'

            args, kwargs = self._parse_value(value,
                                             {"service": service, "config": docker_compose_config},
                                             property_name)
            if event_name == 'docker:binary':
                kwargs["docker_compose_service"] = service_name

            event_parsed_values = parsed_values.get(event_name)
            if not event_parsed_values:
                event_parsed_values = {}
                parsed_values[event_name] = event_parsed_values

            if event_id in event_parsed_values:
                event_args, event_kwargs = event_parsed_values[event_id]
                event_args.extend(args)
                event_kwargs.update(kwargs)
            else:
                event_parsed_values[event_id] = args, kwargs
        return parsed_values

    def _parse_value(self, value, names, property_name=None):
        values = map(str.strip, value.split("|")) if not property_name else [value]

        args = []
        kwargs = {}

        for expression in values:
            if not property_name and "=" in expression:
                var, val = expression.split("=", 1)
            else:
                var, val = property_name, expression

            eval_match = self.eval_re.match(val)
            if eval_match:
                val = simple_eval(eval_match.group(1), names=names)

            if var:
                kwargs[var] = val
            else:
                args.append(val)

        return args, kwargs


class DockerComposeBinaryAction(InitializableAction):
    """
    Convert ddb.event.bus.emit.docker:binary events to binary available from shell.
    """

    def __init__(self):
        super().__init__()
        self.binaries = set()  # type: Set[Binary]

    @property
    def event_bindings(self):
        return (events.docker.binary,
                EventBinding(events.docker.docker_compose_before_events, call=self.before_events),
                EventBinding(events.docker.docker_compose_after_events, call=self.after_events))

    @property
    def name(self) -> str:
        return "docker:docker-compose-binary"

    def initialize(self):
        register_project_cache("docker.binaries")

    def destroy(self):
        if caches.has("docker.binaries"):
            caches.unregister("docker.binaries", callback=lambda c: c.close())

    def before_events(self, docker_compose_config):
        """
        Reset current binaries dict.
        """
        self.binaries = set()

    def after_events(self, docker_compose_config):
        """
        Cache binaries and emit unregistered events for those removed from the previous run.
        """
        docker_binaries_cache = caches.get("docker.binaries")
        cached_binaries = docker_binaries_cache.get("cached_binaries")  # type: Set[Binary]
        if cached_binaries is None:
            cached_binaries = set()

        to_remove_binaries = set()
        for cached_binary in cached_binaries:
            if cached_binary not in self.binaries:
                if binaries.unregister(cached_binary.name, cached_binary):
                    events.binary.unregistered(binary=cached_binary)
                    to_remove_binaries.add(cached_binary)

        for to_remove in to_remove_binaries:
            cached_binaries.remove(to_remove)

        for binary in self.binaries:
            cached_binaries.add(binary)

        docker_binaries_cache.set("cached_binaries", cached_binaries)
        docker_binaries_cache.flush()

    def execute(self, name=None, workdir=None, options=None, options_condition=None, condition=None, args=None,
                exe=False, entrypoint=None, global_=None, docker_compose_service=None):
        """
        Execute action
        """
        if name is None and docker_compose_service:
            name = docker_compose_service
        if name is None:
            raise ValueError("name should be defined")

        binary = DockerBinary(name, docker_compose_service=docker_compose_service, workdir=workdir, options=options,
                              options_condition=options_condition, condition=condition, args=args, exe=exe,
                              entrypoint=entrypoint, global_=global_)

        self.binaries.add(binary)

        if binaries.has(name, binary):
            context.log.notice("Binary exists: %s" % (name,))
            events.binary.found(binary=binary)
            return

        binaries.register(binary)
        context.log.success("Binary registered: %s", name)
        events.binary.registered(binary=binary)


class LocalVolumesAction(Action):
    """
    This should avoid issues where docker creates local volume mount points as root:root.
    We can create those folder before starting the stack with the user account.
    """

    @property
    def event_bindings(self):
        return events.docker.docker_compose_config

    @property
    def name(self) -> str:
        return "docker:docker-compose-local-volumes"

    def execute(self, docker_compose_config: dict):
        """
        Execute action
        """
        if 'services' not in docker_compose_config:
            return

        external_volumes = docker_compose_config['volumes'].keys() if 'volumes' in docker_compose_config else []

        volume_mappings = self._get_volume_mappings(docker_compose_config, external_volumes)

        volume_mappings = [(source, target) for (source, target) in volume_mappings if
                           self._check_file_in_project(source)]

        for source, target in volume_mappings:
            self._create_local_volume(source, target)

        for source_a, destination_a in volume_mappings:
            for source_b, destination_b in volume_mappings:
                if source_a == source_b and destination_a == destination_b:
                    continue

                if destination_b.startswith(destination_a):
                    relative_destination = PurePosixPath(destination_b).relative_to(destination_a)
                    related_path = str(PurePosixPath().joinpath(source_a, relative_destination))
                    rel_related_path = os.path.relpath(os.path.normpath(str(related_path)), ".")

                    if not os.path.exists(rel_related_path):
                        os.makedirs(rel_related_path)
                        context.log.info("Local volume source: %s (related directory created)", rel_related_path)

                    self._fix_owner(rel_related_path)

    @staticmethod
    def _fix_owner(relative_path):
        if hasattr(os, 'chown'):
            stat_info = os.stat(relative_path)
            uid = stat_info.st_uid
            gid = stat_info.st_gid
            if uid != config.data.get("docker.user.uid") and gid != config.data.get("docker.user.gid"):
                context.log.warning("Invalid owner detected for %s", relative_path)
                try:
                    # pylint:disable=no-member
                    os.chown(relative_path, config.data.get("docker.user.uid"), config.data.get("docker.user.gid"))
                    context.log.info("Owner has been fixed for %s", relative_path)
                except OSError:
                    context.log.error("Run this command to fix: sudo chown -R \"%s:%s\" %s",
                                      config.data.get("docker.user.uid"), config.data.get("docker.user.gid"),
                                      relative_path)

    @staticmethod
    def _check_file_in_project(target):
        target_path = os.path.realpath(target)
        cwd = os.path.realpath(".")
        return target_path.startswith(cwd)

    @staticmethod
    def _create_local_volume(source, target):
        rel_source = os.path.relpath(source, ".")
        if os.path.exists(source):
            context.log.notice("Local volume source: %s (exists)", rel_source)
        else:
            _, source_ext = os.path.splitext(source)
            target_ext = None
            if target:
                _, target_ext = os.path.splitext(target)
            if source_ext or target_ext:
                # Create empty file, because with have an extension in source or target.
                os.makedirs(str(Path(source).parent), exist_ok=True)
                context.log.info("Local volume source: %s (file created)", rel_source)
                with open(source, "wb"):
                    pass
                events.file.generated(source=None, target=rel_source)
            else:
                # Create empty directory, because neither source or target has an extension.
                os.makedirs(source)
                context.log.info("Local volume source: %s (directory created)", rel_source)
        LocalVolumesAction._fix_owner(rel_source)

    @staticmethod
    def _get_volume_mappings(docker_compose_config, external_volumes):
        volume_mappings = []
        external_volumes_dict = {}

        for service in docker_compose_config.get('services', {}).values():
            if 'volumes' not in service:
                continue

            for volume_spec in service['volumes']:
                if isinstance(volume_spec, dict):
                    source = volume_spec['source']
                    target = volume_spec['target']
                else:
                    source, target, _ = volume_spec.rsplit(':', 2)

                if source in external_volumes:
                    external_volumes_dict[source] = target
                    continue

                volume_mapping = (source, target)

                if volume_mapping not in volume_mappings:
                    volume_mappings.append(volume_mapping)

        for volume, volume_config in docker_compose_config.get('volumes', {}).items():
            if volume_config and volume_config.get('driver') == 'local' and volume_config.get('driver_opts'):
                driver_opts = volume_config.get('driver_opts')
                device = driver_opts.get('device')
                if device and driver_opts.get('o') == 'bind':
                    source = PurePosixPath(device).joinpath(volume)
                    target = external_volumes_dict.get(volume)
                    if source and target:
                        volume_mapping = (source, target)
                        if volume_mapping not in volume_mappings:
                            volume_mappings.append(volume_mapping)

        return volume_mappings


class DockerDisplayInfoAction(Action):
    """
    Retrieve exposed ports and vhosts from docker-compose config and display them to the user
    """

    def __init__(self):
        super().__init__()
        self.current_yaml_output = None

    @property
    def event_bindings(self):
        return (
            events.phase.info
        )

    @property
    def name(self) -> str:
        return "docker:display-info"

    def execute(self):
        """
        Execute action
        """
        if not os.path.exists("docker-compose.yml"):
            return

        yaml_output = run("docker-compose", "config")

        parsed_config = yaml.load(yaml_output, yaml.SafeLoader)
        docker_compose_config = Dotty(parsed_config)

        if self.current_yaml_output == yaml_output:
            return
        self.current_yaml_output = yaml_output

        services = docker_compose_config.get('services')
        if not services:
            return

        for service_name in sorted(services.keys()):
            service_config = services.get(service_name)
            environments = self._retrieve_environment_data(service_config)
            ports = self._retrieve_service_ports(service_config)
            docker_binaries = self._retrieve_binaries_data(service_config)
            vhosts = self._retrieve_vhosts_data(service_config)

            output = self._output_data(service_name, environments, ports, docker_binaries, vhosts)
            if output:
                print(output)
                print()

        if features.has('traefik'):
            for id_, extra_service_data, _ in TraefikExtraServicesAction.get_extra_services():
                output = self._output_traefik_data(id_, extra_service_data)
                print(output)
                print()

    @staticmethod
    def _retrieve_environment_data(service_config: Dotty) -> Dict[str, str]:
        """
        Retrieve environment data
        :param service_config: the service configuration
        :return: a dict containing environment variables
        """
        environments = service_config.get('environment')
        if not environments:
            return {}

        return environments

    @staticmethod
    def _retrieve_service_ports(service_config: Dotty) -> List[ServicePort]:  # pylint: disable=no-self-use
        """
        Retrieve services ports data
        :param service_config: the service configuration
        :return: a list of service port
        """
        ports = service_config.get('ports')
        if not ports:
            return []

        def _to_service_ports(port):
            if isinstance(port, str):
                return ServicePort.parse(port)

            parameters = {'target': None, 'published': None, 'protocol': None, 'mode': None, 'external_ip': None}
            parameters.update(port)
            return [ServicePort(**parameters)]

        service_ports = []
        for port in ports:
            for service_port in _to_service_ports(port):
                service_ports.append(service_port)

        return service_ports

    @staticmethod
    def _retrieve_binaries_data(service_config: Dotty) -> List[str]:  # pylint: disable=no-self-use
        """
        Retrieve binaries data
        :param service_config: the service configuration
        :return: a list containing binaries
        """
        labels = service_config.get('labels')
        if not labels:
            return []

        binary_regex_re = re.compile(r"^\s*ddb\.emit\.(.+?)(?:\[(.+?)\])?(?:\((.+?)\))?\s*$")

        binaries_labels = []
        for key in labels.keys():
            match = binary_regex_re.match(key)
            if not match:
                continue
            event_name = match.group(1)
            binary_name = match.group(2)
            if event_name == 'docker:binary' and binary_name not in binaries_labels:
                binaries_labels.append(match.group(2))

        return binaries_labels

    @staticmethod
    def _retrieve_vhosts_data(service_config: Dotty) -> List[str]:  # pylint: disable=no-self-use
        """
        Retrieve vhosts data
        :param service_config: the service configuration
        :return: a list containing vhosts data
        """
        labels = service_config.get('labels')
        if not labels:
            return []

        vhosts_regex_re = re.compile(r"^Host\(`(.+?)`\)$")

        vhosts_labels = []
        for key in labels.keys():
            value = labels.get(key)
            match = vhosts_regex_re.match(value)
            if not match:
                continue
            http_url = 'http://{}/'.format(match.group(1))
            https_url = 'https://{}/'.format(match.group(1))

            if '-tls.' in key:
                try:
                    vhosts_labels.remove(http_url)
                except ValueError:
                    pass
                vhosts_labels.append(https_url)
                continue

            if https_url not in vhosts_labels:
                vhosts_labels.append(http_url)

        return vhosts_labels

    @staticmethod
    def _output_data(service_name: str, environments: Dict[str, str],  # pylint: disable=no-self-use
                     ports: List[ServicePort], docker_binaries: List[str], vhosts: List[str]):
        """
        Process the data and render it to the user
        :param service_name: the service name
        :param environments: the service environment data to display
        :param ports: the service ports data to display
        :param docker_binaries: the list of docker binaries
        :param vhosts: the list of vhosts
        :return: a dict containing useful labels data
        """

        header_block = ['{}'.format(service_name)]
        blocks = [header_block]

        if (config.args.type is None or 'env' in config.args.type) and environments:
            tmp_content = []
            for key in sorted(environments.keys()):
                tmp_content.append('{}: {}'.format(key, environments.get(key)))

            blocks.append(tmp_content)

        if (config.args.type is None or 'port' in config.args.type) and ports:
            tmp_content = []
            for port in ports:
                tmp_content.append(port.legacy_repr() if
                                   hasattr(port, 'legacy_repr') else
                                   "%s:%s" % (port.published, port.target))
            blocks.append(tmp_content)

        if (config.args.type is None or 'bin' in config.args.type) and docker_binaries:
            tmp_content = []
            for binary in sorted(docker_binaries):
                tmp_content.append(binary)
            blocks.append(tmp_content)

        if (config.args.type is None or 'vhost' in config.args.type) and vhosts:
            tmp_content = []
            for vhost in sorted(vhosts):
                tmp_content.append(vhost)
            blocks.append(tmp_content)

        if blocks != [header_block]:
            return get_table_display(blocks, False)
        return ''

    @staticmethod
    def _output_traefik_data(id_, extra_service_data):
        if 'domain' in extra_service_data:
            domain = extra_service_data.get('domain')
            if extra_service_data.get('https') in [None, True]:
                domain = 'https://' + domain
            else:
                domain = 'http://' + domain
        else:
            domain = extra_service_data.get('rule')

        blocks = [
            [id_ + " (extra)"],
            [domain + ' --> ' + extra_service_data.get('url')]
        ]
        return get_table_display(blocks, False)



# -*- coding: utf-8 -*-
import os
from typing import Iterable, ClassVar

from dotty_dict import Dotty

from ddb.feature.traefik.schema import ExtraServiceSchema
from .actions import TraefikInstalllCertsAction, TraefikUninstalllCertsAction, TraefikExtraServicesAction
from .schema import TraefikSchema
from ..feature import Feature, FeatureConfigurationAutoConfigureError
from ..schema import FeatureSchema
from ...action import Action
from ...config import config


class TraefikFeature(Feature):
    """
    Traefik support (https://traefik.io).
    """

    @property
    def name(self) -> str:
        return "traefik"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core", "certs[optional]"]

    @property
    def schema(self) -> ClassVar[FeatureSchema]:
        return TraefikSchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            TraefikInstalllCertsAction(),
            TraefikUninstalllCertsAction(),
            TraefikExtraServicesAction()
            # TODO: Add action to install custom traefik configuration
        )

    def _configure_defaults(self, feature_config: Dotty):
        certs_directory = feature_config.get('certs_directory')
        if not certs_directory and config.paths.home:
            certs_directory = os.path.join(config.paths.home, 'certs')
            if os.path.exists(certs_directory):
                feature_config['certs_directory'] = certs_directory

        config_directory = feature_config.get('config_directory')
        if not config_directory and config.paths.home:
            config_directory = os.path.join(config.paths.home, 'traefik', 'config')
            if os.path.exists(config_directory):
                feature_config['config_directory'] = config_directory

            config_directory = os.path.join(config.paths.home, 'docker-toolbox', '.docker', 'traefik', 'hosts')
            if os.path.exists(config_directory):
                feature_config['config_directory'] = config_directory

        extra_services = feature_config.get('extra_services')
        if extra_services:
            for extra_service in extra_services.values():
                domain = extra_service.get('domain')
                if not extra_service.get('rule'):
                    if not domain:
                        raise FeatureConfigurationAutoConfigureError(self, 'extra_services',
                                                                     "domain must be defined when rule is not defined.")
                    extra_service['rule'] = "Host(`%s`)" % domain
                if extra_service.get('https') is not False and not domain:
                    raise FeatureConfigurationAutoConfigureError(self, 'extra_services',
                                                                 "domain must be defined when https is not False.")

        if not feature_config.get('config_directory'):
            feature_config['disabled'] = True



# -*- coding: utf-8 -*-
import pytest
from _pytest.capture import CaptureFixture

from ddb.__main__ import main, ParseCommandLineException


def test_main_no_args(project_loader):
    project_loader("no-config")
    with pytest.raises(ParseCommandLineException) as exc:
        main([])


def test_main_features(project_loader):
    project_loader("no-config")
    main(["features"])


def test_main_config(project_loader):
    project_loader("no-config")
    main(["config"])


def test_main_config_should_display_error(project_loader, capsys: CaptureFixture):
    project_loader("no-config")
    main(["configure"])

    outerr = capsys.readouterr()
    assert not outerr.out
    assert "No project configuration file found." in outerr.err


def test_main_config_variables(project_loader):
    project_loader("no-config")
    main(["config", "--variables"])



import os

import pytest

from ddb.__main__ import load_registered_features
from ddb.config import config
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.utils import file
from ddb.utils.file import FileWalker, FileUtils


class TestHasSameContent:
    def test_should_return_true_if_same_content(self, data_dir: str):
        file.has_same_content(os.path.join(data_dir, "512bytes.bin"), os.path.join(data_dir, "512bytes.copy.bin"))

    def test_should_return_false_if_different_content(self, data_dir: str):
        file.has_same_content(os.path.join(data_dir, "512bytes.bin"), os.path.join(data_dir, "1KB.bin"))

    def test_should_raise_file_not_found_error_if_file_doesnt_exists(self, data_dir: str):
        with pytest.raises(FileNotFoundError):
            file.has_same_content(os.path.join(data_dir, "512bytes.bin"), os.path.join(data_dir, "another.bin"))


class TestFileWalker:
    def test_should_exclude_files_in_excluded_directory(self):
        fw = FileWalker([], ["**/node_modules"], [], ".")
        assert fw.is_source_filtered("blabla/node_modules") is True
        assert fw.is_source_filtered("blabla/node_modules/file") is True
        assert fw.is_source_filtered("blabla/node_modules/subdirectory/file") is True
        assert fw.is_source_filtered("blabla/another/subdirectory/file") is False


class TestFileUtils:
    def test_get_file_content(self, data_dir: str, project_loader):
        project_loader()
        features.register(CoreFeature())
        load_registered_features()

        url = 'https://raw.githubusercontent.com/inetum-orleans/docker-devbox-ddb/b4f11276a37a4e4b1142f6b54b3d0763ccf5639e/ddb/__init__.py'
        path = 'test_file_content.txt'

        expected_file_content = '\n'.join([
            '# -*- coding: utf-8 -*-',
            '',
            'from .__version__ import __version__',
            ''
        ])

        url_content = FileUtils.get_file_content(url)
        assert expected_file_content == url_content

        url_content = FileUtils.get_file_content('file://' + os.path.join(config.path.project_home, path))
        assert url_content == 'this is a file for test_file_content'

        url_content = FileUtils.get_file_content('file://' + path)
        assert url_content == 'this is a file for test_file_content'



import re

from _pytest.logging import LogCaptureFixture

from ddb.config import migrations
from ddb.config.migrations import PropertyMigration, MigrationsDotty


class TestWarnDeprecatedDotty:
    def teardown_method(self, test_method):
        migrations.set_history()

    def test_warn_deprecated_simple(self, caplog: LogCaptureFixture):
        history = (
            PropertyMigration("old_property",
                              "new_property", since="v1.1.0"),
        )

        migrations.set_history(history)

        data = MigrationsDotty({
            "old_property": "value"
        })

        migrations.migrate(data)

        assert data.get("old_property") == "value"
        assert data.get("new_property") == "value"

        logs = caplog.text

        assert re.match(
            r"WARNING .*\"old_property\" configuration property is deprecated since v1\.1\.0 and will be removed "
            r"in a future major release\. You should use \"new_property\" instead\.", logs)

    def test_warn_deprecated_deep(self, caplog: LogCaptureFixture):
        history = (
            PropertyMigration("some.namespace.old_property",
                              "another.namespace.new_property"),
        )

        migrations.set_history(history)

        data = MigrationsDotty()
        data["some.namespace.old_property"] = "value"

        migrations.migrate(data)

        assert data.get("some.namespace.old_property") == "value"
        assert data.get("another.namespace.new_property") == "value"

        assert data.get("some").get("namespace").get("old_property") == "value"
        assert data.get("another").get("namespace").get("new_property") == "value"

        assert data.get("some.namespace").get("old_property") == "value"
        assert data.get("another.namespace").get("new_property") == "value"

        logs = caplog.text

        assert re.match(
            r"WARNING .*\"some.namespace.old_property\" configuration property is deprecated and will be removed "
            r"in a future major release\. You should use \"another.namespace.new_property\" instead\.", logs)

    def test_compat_deep(self, caplog: LogCaptureFixture):
        history = (
            PropertyMigration("some.namespace.old_property",
                              "another.namespace.new_property"),
        )

        migrations.set_history(history)

        data = MigrationsDotty()
        data["another.namespace.new_property"] = "value"

        assert data.get("some.namespace.old_property") == "value"
        assert data.get("another.namespace.new_property") == "value"

        assert data.get("some").get("namespace").get("old_property") == "value"
        assert data.get("another").get("namespace").get("new_property") == "value"

        assert data.get("some.namespace").get("old_property") == "value"
        assert data.get("another.namespace").get("new_property") == "value"

        logs = caplog.text

        assert re.match(
            r"WARNING .*\"some.namespace.old_property\" configuration property is deprecated and will be removed "
            r"in a future major release\. You should use \"another.namespace.new_property\" instead\.", logs)

    def test_compat_deep_merge(self, caplog: LogCaptureFixture):
        history = (
            PropertyMigration("some.namespace.old_property",
                              "another.namespace.new_property"),
        )

        migrations.set_history(history)

        data = MigrationsDotty()
        data["another.namespace.new_property"] = "value"
        data["some.namespace.another"] = "foo"

        assert data.get("some.namespace.another") == "foo"
        assert data.get("some.namespace.old_property") == "value"
        assert data.get("another.namespace.new_property") == "value"

        assert data.get("some").get("namespace").get("another") == "foo"
        assert data.get("some").get("namespace").get("old_property") == "value"
        assert data.get("another").get("namespace").get("new_property") == "value"

        logs = caplog.text

        assert re.match(
            r"WARNING .*\"some.namespace.old_property\" configuration property is deprecated and will be removed "
            r"in a future major release\. You should use \"another.namespace.new_property\" instead\.", logs)



import os

import pytest
from _pytest.capture import CaptureFixture

from ddb.__main__ import load_registered_features, register_actions_in_event_bus
from ddb.binary import binaries
from ddb.binary.binary import DefaultBinary
from ddb.config import config
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.feature.docker.binaries import DockerBinary
from ddb.feature.run import RunFeature, RunAction
from ddb.feature.shell import ShellFeature
from ddb.utils.docker import DockerUtils

docker_compose_bin = "docker-compose" if os.name != "nt" else "docker-compose.exe"


class TestRunFeature:
    def test_empty_project_without_core(self, project_loader):
        project_loader("empty")

        features.register(RunFeature())
        features.register(ShellFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = RunAction()
        action.execute()

    def test_empty_project_with_core(self, project_loader):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(RunFeature())
        features.register(ShellFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = RunAction()
        action.execute()

    def test_run_missing_binary(self, project_loader):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(RunFeature())
        features.register(ShellFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = RunAction()

        with pytest.raises(ValueError):
            action.run("missing")

    def test_run_existing_binary(self, project_loader, capsys: CaptureFixture):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(RunFeature())
        features.register(ShellFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        binaries.register(DefaultBinary("test", ["some", "command"]))

        action = RunAction()
        action.run("test")

        read = capsys.readouterr()

        assert read.out == "some command\n"

    def test_run_docker_binary(self, project_loader, capsys: CaptureFixture):
        project_loader("empty")
        config.cwd = config.paths.project_home

        features.register(CoreFeature())
        features.register(RunFeature())
        features.register(ShellFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        binaries.register(DockerBinary("bin", "service", "/app"))

        action = RunAction()
        action.run("bin")

        read = capsys.readouterr()

        assert read.out == docker_compose_bin + " run --rm --workdir=/app/. service\n"

    def test_run_docker_binary_workdir(self, project_loader, capsys: CaptureFixture):
        project_loader("empty")
        config.cwd = os.path.join(config.paths.project_home, "sub")

        features.register(CoreFeature())
        features.register(RunFeature())
        features.register(ShellFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        binaries.register(DockerBinary("bin", "service", "/app"))

        action = RunAction()
        action.run("bin")

        read = capsys.readouterr()

        assert read.out == docker_compose_bin + " run --rm --workdir=/app/sub service\n"

    def test_run_docker_binary_workdir_outside(self, project_loader, capsys: CaptureFixture):
        project_loader("outside")
        config.cwd = os.path.join(config.paths.project_home, "../outside")

        features.register(CoreFeature())
        features.register(RunFeature())
        features.register(ShellFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        binaries.register(DockerBinary("bin", "service", "/app"))

        action = RunAction()
        action.run("bin")

        read = capsys.readouterr()

        cwd = config.cwd if config.cwd else os.getcwd()
        real_cwd = os.path.realpath(cwd)

        assert read.out == docker_compose_bin + " -f ../project/docker-compose.yml " \
                                                "run --rm "\
                                                f"--volume={real_cwd}:/app " \
                                                "--workdir=/app " \
                                                "service\n"

    def test_run_docker_binary_exe(self, project_loader, capsys: CaptureFixture):
        project_loader("exe")

        features.register(CoreFeature())
        features.register(RunFeature())
        features.register(ShellFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        binaries.register(DockerBinary("echo",
                                       docker_compose_service="web",
                                       args="echo",
                                       exe=True))

        DockerUtils.service_stop('web')
        assert not DockerUtils.is_container_up('web')

        action = RunAction()
        action.run("echo")

        read = capsys.readouterr()

        assert read.out == docker_compose_bin + " exec web echo\n"

        assert DockerUtils.is_container_up('web')

    def test_run_binary_condition(self, project_loader, capsys: CaptureFixture):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(RunFeature())
        features.register(ShellFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        binary1 = DockerBinary("npm", "node1")
        binary2 = DockerBinary("npm", "node2")
        default_binary = DefaultBinary("npm", ["npm"])
        binary4 = DockerBinary("npm", "node4")
        binary_invalid_condition = DockerBinary("npm", "node5", condition="'another-value' in args")
        binary_valid_condition = DockerBinary("npm", "node6", condition="'some-value' in args")

        for bin in (binary1, binary2, default_binary, binary4, binary_invalid_condition, binary_valid_condition):
            binaries.register(bin)

        action = RunAction()
        action.run("npm", "some-value")

        read = capsys.readouterr()
        assert "node6" in read.out



# -*- coding: utf-8 -*-
from abc import ABC
from os import linesep
from typing import ClassVar, Iterable, Union

from dotty_dict import Dotty, dotty
from marshmallow import ValidationError

from .schema import FeatureSchema
from ..action import Action
from ..binary import Binary
from ..command import Command
from ..config import config
from ..phase import Phase
from ..registry import RegistryObject
from ..service import Service


class Feature(RegistryObject, ABC):  # pylint:disable=abstract-method
    """
    A feature provides phases, commands, binaries, actions and services. It can be configured with a key matching it's
    name in the configuration, and it's configuration should match a given schema.
    """

    @property
    def schema(self) -> ClassVar[FeatureSchema]:
        """
        Configuration schema for this feature.
        """
        return FeatureSchema

    @property
    def phases(self) -> Iterable[Phase]:
        """
        Phases provided by this feature.
        """
        return []

    @property
    def commands(self) -> Iterable[Command]:
        """
        Commands provided by this feature.
        """
        return []

    @property
    def binaries(self) -> Iterable[Binary]:
        """
        Binaries provided by this feature.
        """
        return []

    @property
    def actions(self) -> Iterable[Action]:
        """
        Actions provided by this feature.
        """
        return []

    @property
    def services(self) -> Iterable[Service]:
        """
        Services provided by this feature.
        """
        return []

    @property
    def dependencies(self) -> Iterable[str]:
        """
        Feature dependencies. It should match the name of other features. If may ends with '[optional]' if the
        dependency is optional.
        """
        return []

    def configure(self, bootstrap=False):
        """
        Configure this feature.
        """
        feature_config = config.data.get(self.name)
        try:
            valid_feature_config = self.validate_feature_configuration(feature_config, bootstrap)
        except ValidationError as err:
            raise FeatureConfigurationValidationError(self, err) from err
        config.data[self.name] = valid_feature_config

    def validate_feature_configuration(self, feature_config: Dotty, bootstrap=False):
        """
        Sanitize and validate using given schema part of the configuration.
        """
        schema = self.schema()
        raw_feature_config = feature_config

        if not raw_feature_config:
            raw_feature_config = {}
        feature_config = schema.dump(raw_feature_config)

        if not bootstrap:
            self._configure_defaults(dotty(feature_config))

            feature_config = schema.load(feature_config)
        feature_config = config.apply_environ_overrides(feature_config, config.env_override_prefix + "_" + self.name)

        return feature_config

    def _configure_defaults(self, feature_config: Dotty):
        """
        Override this method to load default values that may depend on the context.
        """

    def before_load(self):
        """
        Invoked before this feature is loaded.
        """

    def after_load(self):
        """
        Invoked after this feature is loaded.
        """

    @property
    def disabled(self):
        """
        Check if this feature is disabled.
        """
        return config.data.get(self.name + '.disabled')


class FeatureConfigurationError(Exception):
    """
    A generic feature configuration error.
    """

    def __init__(self, feature: Feature, *additional_messages: str):
        message = "Feature \"" + feature.name + "\" has invalid configuration."
        for additional_message in additional_messages:
            message += linesep + " " * 2 + additional_message
        super().__init__(message)


class FeatureConfigurationAutoConfigureError(FeatureConfigurationError):
    """
    Feature configuration error raised when something wrongs occurs during auto-configuration.
    """

    def __init__(self, feature: Feature, setting_name: str = None, error: Union[Exception, str] = None):
        super().__init__(feature,
                         setting_name + ": " + "auto-configuration has failed" +
                         (" (" + str(error) + "). " if error else ". ") +
                         "Please configure " + feature.name + "." + setting_name + " manually.")


class FeatureConfigurationReadOnlyError(FeatureConfigurationError):
    """
    Feature configuration error raised when a read-only property is defined in configuration files user.
    """

    def __init__(self, feature: Feature, setting_name: str = None):
        super().__init__(feature,
                         setting_name + ": " + "auto-configuration has failed" +
                         feature.name + "." + setting_name + " is a read-only property computed from other properties "
                                                             "and should not be defined in configuration files.")



class FeatureConfigurationValidationError(FeatureConfigurationError):
    """
    Feature configuration error raised on validation error.
    """

    def __init__(self, feature: Feature, validation_error: ValidationError = None):
        additional_messages = []
        for (field, field_messages) in validation_error.messages.items():
            for field_message in field_messages:
                additional_messages.append("%s: %s" % (field, field_message))
        super().__init__(feature, *additional_messages)
        self.validation_error = validation_error



import os
import shlex
import subprocess
from abc import ABC, abstractmethod
from typing import Tuple, Iterable, Dict, Any, Optional

from slugify import slugify

from ddb.binary import Binary
from ddb.config import config
from ddb.utils.file import force_remove, write_if_different, chmod


class ShellIntegration(ABC):
    """
    Interface for shell integration.
    """

    def __init__(self, name, description):
        self.name = name
        self.description = description

    @abstractmethod
    def set_environment_variable(self, key, value) -> Iterable[str]:
        """
        Returns an instruction that set an environment variable to shell.
        """

    @abstractmethod
    def remove_environment_variable(self, key) -> Iterable[str]:
        """
        Returns an instruction that unset an environment variable to shell.
        """

    @abstractmethod
    def remove_binary_shim(self, shims_path: str, name: str) -> Optional[str]:
        """
        Delete a binary shim for this shell.
        @return removed filepath
        """

    @abstractmethod
    def create_binary_shim(self, shims_path: str, name: str, global_: bool) -> Tuple[bool, str]:
        """
        Add a binary shim for this shell.
        :return created filepath
        """

    @abstractmethod
    def create_alias_binary_shim(self, shims_path: str, binary: Binary) -> Tuple[bool, str]:
        """
        Add a alias binary shim for this shell.
        :return created filepath
        """

    @abstractmethod
    def evaluate_script(self, script_filepath) -> Iterable[str]:
        """
        Get the command to evaluate the script inside the current shell context.
        """

    @abstractmethod
    def generate_and_operator(self, new_line: bool = False) -> str:
        """
        Build the shell command to run the given command inside the current shell context.
        """

    @abstractmethod
    def generate_or_operator(self, new_line: bool = False) -> str:
        """
        Build the shell command to run the given command inside the current shell context.
        """

    @abstractmethod
    def generate_cmdline(self, command: Iterable[str], system_path=True) -> str:
        """
        Build the shell command to run the given command inside the current shell context.
        """

    def before_environ_backup(self, environ: Dict[str, str]):
        """
        Alter environment variables before backup. This should be used to remove for saved environ variables that
        should not be restored on deactivation.
        """

    def header(self) -> Iterable[str]:  # pylint:disable=no-self-use
        """
        Returns header of script
        """
        yield from ()

    def footer(self) -> Iterable[str]:  # pylint:disable=no-self-use
        """
        Returns footer of script
        """
        yield from ()

    @property
    def temporary_file_kwargs(self) -> Dict[str, Any]:
        """
        Additional options for temporary file
        """
        return {}


class BashShellIntegration(ShellIntegration):
    """
    Bash integration.
    """

    def __init__(self):
        super().__init__("bash", "Bash")

    @staticmethod
    def _sanitize_key(key):
        return slugify(key, regex_pattern=r'[^-a-zA-Z0-9_]+', separator="_").upper()

    def set_environment_variable(self, key, value):
        yield "export " + self._sanitize_key(key) + "=" + shlex.quote(value)

    def remove_environment_variable(self, key):
        yield "unset " + self._sanitize_key(key)

    def remove_binary_shim(self, shims_path: str, name: str) -> Optional[str]:
        shim = os.path.join(shims_path, name)
        if not os.path.isfile(shim):
            return None
        force_remove(shim)
        return shim

    def create_binary_shim(self, shims_path: str, name: str, global_: bool):
        command = []
        if global_:
            ddb_project_home_variable = next(
                self.set_environment_variable(config.env_prefix + "_PROJECT_HOME", config.paths.project_home)
            )

            compose_ignore_orphans_variable = next(
                self.set_environment_variable("COMPOSE_IGNORE_ORPHANS", "1")
            )

            command.extend(
                ("$(ddb deactivate --force)",
                 f"{ddb_project_home_variable}",
                 f"{compose_ignore_orphans_variable}",
                 "$(ddb activate --force)")
            )

        command.extend(
            ('shim=$(tempfile -p ddb.run.)',
             'echo "rm $shim">>"$shim"',
             f'echo $(ddb run {name} \"$@\") \'\"$@\"\'>>"$shim"',
             'source "$shim"')
        )

        return self._write_shim(shims_path, name, 'binary', command)

    def create_alias_binary_shim(self, shims_path: str, binary: Binary) -> Tuple[bool, str]:
        return self._write_shim(shims_path, binary.name, 'alias', [' '.join(binary.command()) + ' "$@"'])

    @staticmethod
    def _write_shim(shims_path: str, shim_name: str, shim_type: str, lines: Iterable[str]) -> Tuple[bool, str]:
        """
        Generate the shim file
        :return:
        """
        os.makedirs(shims_path, exist_ok=True)
        shim = os.path.join(os.path.normpath(shims_path), shim_name)

        all_lines = ["#!/usr/bin/env bash", "# ddb:shim:" + shim_type]
        all_lines.extend(lines)

        content = '\n'.join(all_lines) + "\n"

        written = write_if_different(shim, content, newline="\n")

        chmod(shim, '+x', logging=False)
        return written, shim

    @property
    def temporary_file_kwargs(self) -> Dict[str, Any]:
        return {"encoding": "utf-8", "newline": '\n'}

    def evaluate_script(self, script_filepath) -> Iterable[str]:
        yield ". %s" % (script_filepath,)

    def generate_and_operator(self, new_line: bool = False) -> str:
        return " &&" + ("\\" if new_line else " ")

    def generate_or_operator(self, new_line: bool = False) -> str:
        return " ||" + ("\\" if new_line else " ")

    def generate_cmdline(self, command: Iterable[str], system_path=True) -> str:
        return subprocess.list2cmdline(command)


class CmdShellIntegration(ShellIntegration):
    """
    Windows cmd integration
    """

    def __init__(self):
        super().__init__("cmd", "Windows cmd.exe")

    @staticmethod
    def _sanitize_key(key):
        return slugify(key, regex_pattern=r'[^-a-zA-Z0-9_]+', separator="_").upper()

    def set_environment_variable(self, key, value):
        # TODO: Maybe use subprocess.list2cmdline for Windows ?
        yield "set " + self._sanitize_key(key) + "=" + value.replace('\n', '!NL! ^\n')

    def remove_environment_variable(self, key):
        yield "set " + self._sanitize_key(key) + "="

    def remove_binary_shim(self, shims_path: str, name: str) -> Optional[str]:
        shim = os.path.join(shims_path, name + '.bat')
        if not os.path.isfile(shim):
            return None
        force_remove(shim)
        return shim

    def create_binary_shim(self, shims_path: str, name: str, global_: bool):
        commands = []
        if global_:
            commands.extend(
                ("set command=(ddb deactivate --force)",
                 "%command%>cmd.txt",
                 "set /p execution=<cmd.txt",
                 "del cmd.txt",
                 "%execution%")
            )

            commands.append(next(
                self.set_environment_variable(config.env_prefix + "_PROJECT_HOME", config.paths.project_home)
            ))

            commands.append(next(
                self.set_environment_variable("COMPOSE_IGNORE_ORPHANS", "1")
            ))

            commands.extend(
                ("set command=(ddb activate --force)",
                 "%command%>cmd.txt",
                 "set /p execution=<cmd.txt",
                 "del cmd.txt",
                 "%execution%")
            )

        commands.extend(
            (f"set command=(ddb run {name} \"%*\")",
             "%command%>cmd.txt",
             "set /p execution=<cmd.txt",
             "del cmd.txt",
             "%execution% \"%*\"")
        )

        return self._write_shim(shims_path, name, 'binary', commands)

    def create_alias_binary_shim(self, shims_path: str, binary: Binary):
        return self._write_shim(shims_path, binary.name, 'alias', [' '.join(binary.command())])

    @staticmethod
    def _write_shim(shims_path: str, shim_name: str, shim_type: str, lines: Iterable[str]) -> Tuple[bool, str]:
        """
        Generate the shim file
        :return:
        """
        os.makedirs(shims_path, exist_ok=True)
        shim = os.path.join(os.path.normpath(shims_path), shim_name + '.bat')

        all_lines = ["@echo off", "REM ddb:shim:" + shim_type, "SHIFT"]
        all_lines.extend(lines)

        content = '\n'.join(all_lines) + "\n"

        written = write_if_different(shim, content, newline="\n")

        chmod(shim, '+x', logging=False)
        return written, shim

    def evaluate_script(self, script_filepath) -> Iterable[str]:
        yield "call %s" % (script_filepath,)

    def header(self) -> Iterable[str]:
        return ["@echo off", "set NL=^", ""]

    def generate_and_operator(self, new_line: bool = False) -> str:
        return " &&" + ("\\" if new_line else " ")

    def generate_or_operator(self, new_line: bool = False) -> str:
        return " ||" + ("\\" if new_line else " ")

    def generate_cmdline(self, command: Iterable[str], system_path=True) -> str:
        return subprocess.list2cmdline(command)



from ddb.config import config


def get_mapped_path(path: str):
    """
    Get docker mapped path, using `docker.path_mapping` configuration.
    :param path:
    :return:
    """
    path_mapping = config.data.get('docker.path_mapping')
    fixed_path = None
    fixed_source_path = None
    if path_mapping:
        for source_path, target_path in path_mapping.items():
            if path.startswith(source_path) and \
                    (not fixed_source_path or len(source_path) > len(fixed_source_path)):
                fixed_source_path = source_path
                fixed_path = target_path + path[len(source_path):]
    return fixed_path if fixed_path else path



import os
import re
import time

from cfssl import cfssl

from ddb.config import config
from ddb.config.config import ConfigPaths
from ddb.feature.bootstrap import bootstrap_features_configuration
from ddb.feature.gitignore import UpdateGitignoreAction


def load_config(data_dir: str = None, name: str = None):
    init_config_paths(data_dir, name)

    bootstrap_features_configuration()
    config.load(config.data)

    return config


def init_config_paths(data_dir: str = None, name: str = None):
    root_dir = os.path.join(data_dir, name) if name else data_dir

    paths = ConfigPaths(ddb_home=os.path.join(root_dir, 'ddb_home'), home=os.path.join(root_dir, 'home'),
                        project_home=os.path.join(root_dir, 'project'))

    if not [path for path in paths if os.path.isdir(path)]:
        paths = ConfigPaths(ddb_home=None, home=None, project_home=root_dir)

    config.paths = paths

    return config


def _get_docker_ip():
    docker_host = os.environ.get('DOCKER_HOST')
    if not docker_host:
        return '127.0.0.1'

    match = re.match(r"tcp:\/\/(.*?):(.*)", docker_host)
    if match:
        return match.group(1)


def _wait_cfssl_ready(max_retry=20):
    client = cfssl.CFSSL(**config.data['certs.cfssl.server'])
    retry = 0
    while True:
        try:
            client.info("")
            break
        except:
            retry += 1
            time.sleep(1)
            if retry > max_retry:
                raise


def setup_cfssl(container_getter):
    cfssl_service = container_getter.get('cfssl')

    config.data['certs.cfssl.server.host'] = _get_docker_ip()
    config.data['certs.cfssl.server.port'] = int(cfssl_service.network_info[0].host_port)
    config.data['certs.cfssl.server.ssl'] = False

    config.defaults.update(config.data)

    _wait_cfssl_ready()


def expect_gitignore(gitignore: str, *expected_lines: str):
    in_block_lines = set()

    if os.path.exists(gitignore):
        with open(gitignore, 'r') as file:
            inside_block = False
            for gitignore_line in file.read().splitlines():
                if UpdateGitignoreAction._is_block_limit(gitignore_line, True):
                    inside_block = True
                    continue
                if UpdateGitignoreAction._is_block_limit(gitignore_line, False):
                    inside_block = False
                    continue
                if inside_block:
                    in_block_lines.add(gitignore_line)

    for expected_line in expected_lines:
        if expected_line not in in_block_lines:
            return False

    return True


def get_user_uid_gid(username):
    try:
        import pwd  # pylint:disable=import-outside-toplevel
        user = pwd.getpwnam(username)
        return user.pw_uid, user.pw_gid
    except ImportError:
        return ''


def get_group_gid(groupname):
    try:
        import grp  # pylint:disable=import-outside-toplevel
        try :
            group = grp.getgrnam(groupname)
            return group.gr_gid
        except KeyError:
            pass

        import pwd  # pylint:disable=import-outside-toplevel
        user = pwd.getpwnam(groupname)
        return user.pw_gid
    except ImportError:
        return ''



# -*- coding: utf-8 -*-
from marshmallow import fields

from ddb.feature.schema import FeatureSchema


class FileSchema(FeatureSchema):
    """
    File schema.
    """
    suffixes = fields.List(fields.String())
    extensions = fields.List(fields.String())
    includes = fields.List(fields.String())
    excludes = fields.List(fields.String(), default=["**/_*",
                                                     "**/.git",
                                                     "**/.idea",
                                                     "**/node_modules",
                                                     "**/vendor",
                                                     "**/target",
                                                     "**/dist"])



# -*- coding: utf-8 -*-
import glob
import os
import re

import requests

from ddb.action import Action
from ddb.cache import caches, requests_cache_name
from ddb.config import config
from ddb.event import events
from ddb.utils.file import write_if_different, copy_if_different


def copy_from_url(source, destination, filename=None):
    """
    Copy from an URL source.
    """
    cache = caches.get(requests_cache_name)
    response = cache.get(source)
    if not response:
        response = requests.get(source, allow_redirects=True)
        response.raise_for_status()
        cache.set(source, response)
    if not filename:
        content_disposition = response.headers['content-disposition']
        filename = re.findall("filename=(.+)", content_disposition)[0]
    target_path = os.path.join(destination, filename)
    if write_if_different(target_path, response.content, 'rb', 'wb', log_source=source) or config.eject:
        return target_path
    return None


def get_dispatch_directories(dispatch):
    """
    Get dispatch directories from dispatch glob
    """
    dispatch_directories = []
    if dispatch:
        for dispatch_expr in dispatch:
            if dispatch_expr.startswith("!"):
                matches = glob.glob(dispatch_expr[1:])
                dirs = [match for match in matches if os.path.isdir(match)]
                for to_remove in dirs:
                    try:
                        dispatch_directories.remove(to_remove)
                    except ValueError:
                        pass
            else:
                matches = glob.glob(dispatch_expr)
                dirs = [match for match in matches if os.path.isdir(match)]
                for to_add in dirs:
                    if to_add not in dispatch_directories:
                        dispatch_directories.append(to_add)

    return dispatch_directories


class CopyAction(Action):
    """
    Copy files from local filesystem or URL to one of many directories.
    """

    @property
    def name(self) -> str:
        return "copy:copy"

    @property
    def event_bindings(self):
        return events.phase.configure

    @staticmethod
    def execute():
        """
        Execute action
        """
        specs = config.data.get("copy.specs")
        if not specs:
            return

        generated_events = []

        for spec in specs:
            source = spec['source']
            destination = spec.get('destination')

            dispatch = spec.get('dispatch')
            dispatch_directories = get_dispatch_directories(dispatch)

            if not dispatch_directories:
                dispatch_directories = ['']

            for dispatch_directory in dispatch_directories:
                if destination:
                    file_destination = os.path.relpath(os.path.join(dispatch_directory, destination))
                    os.makedirs(file_destination, exist_ok=True)
                else:
                    file_destination = os.path.relpath(dispatch_directory)

                for event_source, event_target in CopyAction._copy_from_spec(file_destination, source, spec):
                    generated_events.append((event_source, event_target))

        for event_source, event_target in generated_events:
            events.file.generated(source=event_source, target=event_target)

    @staticmethod
    def _copy_from_spec(destination, source, spec):
        if source.startswith('http://') or source.startswith('https://'):
            target_path = copy_from_url(source, destination, spec.get('filename'))
            if target_path:
                yield None, target_path
        elif os.path.exists(source) and os.path.isfile(source):
            filename = spec.get('filename', os.path.basename(source))
            target_path = os.path.join(destination, filename)
            if copy_if_different(source, target_path, 'rb', 'wb', log=True) or config.eject:
                yield source, target_path
        else:
            if os.path.isdir(source):
                source = source + '/*'
            for file in glob.glob(source):
                target_path = os.path.join(destination, os.path.basename(file))
                if copy_if_different(file, target_path, 'rb', 'wb', log=True) or config.eject:
                    yield file, target_path



from pytest_mock import MockerFixture

from ddb.__main__ import main, load_registered_features
from ddb.config import config
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.feature.version import VersionFeature, is_git_repository


class TestVersionFeature:
    def test_empty_project_without_core(self, project_loader, mocker: MockerFixture):
        mocker.patch('ddb.feature.version.is_git_repository', is_git_repository)

        project_loader("no_repo")

        features.register(VersionFeature())
        load_registered_features()

    def test_empty_project_with_core(self, project_loader, mocker: MockerFixture):
        mocker.patch('ddb.feature.version.is_git_repository', is_git_repository)

        project_loader("no_repo")

        features.register(CoreFeature())
        features.register(VersionFeature())
        load_registered_features()

    def test_no_tag_repo(self, project_loader, mocker: MockerFixture):
        mocker.patch('ddb.feature.version.is_git_repository', is_git_repository)

        project_loader("no_tag_repo")

        main(["configure"], reset_disabled=True)

        assert config.data.get('version.hash') is not None
        assert config.data.get('version.short_hash') is not None
        assert config.data.get('version.branch') == 'master'
        assert config.data.get('version.version') is None
        assert config.data.get('version.tag') is None

    def test_tag_repo(self, project_loader, mocker: MockerFixture):
        mocker.patch('ddb.feature.version.is_git_repository', is_git_repository)

        project_loader("tag_repo")

        main(["configure"], reset_disabled=True)

        assert config.data.get('version.hash') is not None
        assert config.data.get('version.short_hash') is not None
        assert config.data.get('version.branch') == 'master'
        assert config.data.get('version.version') == 'v1.0.0'
        assert config.data.get('version.tag') == 'v1.0.0'

    def test_branch_repo(self, project_loader, mocker: MockerFixture):
        mocker.patch('ddb.feature.version.is_git_repository', is_git_repository)

        project_loader("branch_repo")

        main(["configure"], reset_disabled=True)

        assert config.data.get('version.hash') is not None
        assert config.data.get('version.short_hash') is not None
        assert config.data.get('version.branch') == 'some-branch'
        assert config.data.get('version.version') is None
        assert config.data.get('version.tag') is None

    def test_detached_repo(self, project_loader, mocker: MockerFixture):
        mocker.patch('ddb.feature.version.is_git_repository', is_git_repository)

        project_loader("detached_repo")

        main(["configure"], reset_disabled=True)

        assert config.data.get('version.hash') is not None
        assert config.data.get('version.short_hash') is not None
        assert config.data.get('version.branch') == 'develop'
        assert config.data.get('version.version') is None
        assert config.data.get('version.tag') is None

    def test_no_repo(self, project_loader, mocker: MockerFixture):
        mocker.patch('ddb.feature.version.is_git_repository', is_git_repository)

        project_loader("no_repo")

        main(["configure"], reset_disabled=True)

        assert config.data.get('version.hash') is None
        assert config.data.get('version.short_hash') is None
        assert config.data.get('version.branch') is None
        assert config.data.get('version.version') is None
        assert config.data.get('version.tag') is None

    def test_repo_no_commit(self, project_loader, mocker: MockerFixture):
        mocker.patch('ddb.feature.version.is_git_repository', is_git_repository)

        project_loader("repo_no_commit")

        main(["configure"], reset_disabled=True)

        assert config.data.get('version.hash') is None
        assert config.data.get('version.short_hash') is None
        assert config.data.get('version.branch') is None
        assert config.data.get('version.version') is None
        assert config.data.get('version.tag') is None



import os
import shutil

import pytest

from _pytest.capture import CaptureFixture
from compose.config.types import ServicePort
from dotty_dict import Dotty

from ddb.__main__ import load_registered_features, register_actions_in_event_bus
from ddb.action import actions
from ddb.binary import binaries
from ddb.binary.binary import DefaultBinary
from ddb.config import config
from ddb.event import bus
from ddb.feature import features
from ddb.feature.certs import CertsFeature
from ddb.feature.core import CoreFeature
from ddb.feature.docker import DockerDisplayInfoAction
from ddb.feature.docker import DockerFeature, EmitDockerComposeConfigAction
from ddb.feature.docker.binaries import DockerBinary
from ddb.feature.docker.utils import get_mapped_path
from ddb.feature.traefik import TraefikFeature
from ddb.utils.process import effective_command
from tests.utilstest import setup_cfssl


class TestDockerFeature:
    def test_empty_project_without_core(self, project_loader):
        project_loader("empty")

        features.register(DockerFeature())
        load_registered_features()

        action = actions.get('docker:emit-docker-compose-config')  # type:EmitDockerComposeConfigAction
        action.execute()

    def test_empty_project_with_core(self, project_loader):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(DockerFeature())
        load_registered_features()

        action = actions.get('docker:emit-docker-compose-config')  # type:EmitDockerComposeConfigAction
        action.execute()

    def test_ubuntu(self, project_loader):
        project_loader("ubuntu")

        features.register(DockerFeature())
        load_registered_features()

        custom_event_listeners = []

        def listener(docker_compose_config):
            custom_event_listeners.append(docker_compose_config)

        bus.on("docker:docker-compose-config", listener)

        action = actions.get('docker:emit-docker-compose-config')  # type:EmitDockerComposeConfigAction
        action.execute()

        assert len(custom_event_listeners) == 1
        assert custom_event_listeners[0] in [
            # docker-compose >= 1.27.0 returns major version only. see https://github.com/docker/compose/issues/7730
            {"version": "3", "services": {"docker": {"image": "ubuntu"}}},
            {"version": "3.7", "services": {"docker": {"image": "ubuntu"}}}
        ]

    def test_emit_one_arg(self, project_loader):
        project_loader("emit-one-arg")

        features.register(DockerFeature())
        load_registered_features()

        custom_event_listeners = []

        def listener(data):
            custom_event_listeners.append(data)

        bus.on("some:test", listener)

        action = actions.get('docker:emit-docker-compose-config')  # type:EmitDockerComposeConfigAction
        action.execute()

        assert len(custom_event_listeners) == 1
        assert custom_event_listeners[0] == "emit-one-arg"

    def test_emit_one_arg_eval(self, project_loader):
        project_loader("emit-one-arg-eval")

        features.register(DockerFeature())
        load_registered_features()

        custom_event_listeners = []

        def listener(data):
            custom_event_listeners.append(data)

        bus.on("some:test", listener)

        action = actions.get('docker:emit-docker-compose-config')  # type:EmitDockerComposeConfigAction
        action.execute()

        assert len(custom_event_listeners) == 1
        assert custom_event_listeners[0] == 3

    def test_emit_complex(self, project_loader):
        project_loader("emit-complex")

        features.register(DockerFeature())
        load_registered_features()

        some_events = []
        another_events = []

        def someListener(*args, **kwargs):
            some_events.append({"args": args, "kwargs": kwargs})

        def anotherListener(*args, **kwargs):
            another_events.append({"args": args, "kwargs": kwargs})

        bus.on("some:test", someListener)
        bus.on("another:test", anotherListener)

        action = actions.get('docker:emit-docker-compose-config')  # type:EmitDockerComposeConfigAction
        action.execute()

        assert len(some_events) == 3
        assert {"args": ("emit-one-arg",), "kwargs": {}} in some_events
        assert {"args": ("emit-one-arg-2",), "kwargs": {}} in some_events
        # docker-compose >= 1.27.0 returns major version only. see https://github.com/docker/compose/issues/7730
        assert {"args": ("emit-some-arg",),
                "kwargs": {'image': 'ubuntu', 'kw1': 'emit-one-kwarg', 'kw2': 7, 'version': '3.7'}} in some_events or \
               {"args": ("emit-some-arg",),
                "kwargs": {'image': 'ubuntu', 'kw1': 'emit-one-kwarg', 'kw2': 7, 'version': '3'}} in some_events

        assert len(another_events) == 2
        assert {"args": ("emit-another-arg",), "kwargs": {}} in another_events
        assert {"args": (),
                "kwargs": {"kw1": "emit-another-kwarg"}} in another_events

    def test_binary_workdir(self, project_loader):
        project_loader("binary-workdir")

        features.register(DockerFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = actions.get('docker:emit-docker-compose-config')  # type:EmitDockerComposeConfigAction
        action.execute()

        assert len(list(binaries.all())) == 2
        assert binaries.has("npm")
        assert binaries.has("node")

    def test_binary_options(self, project_loader):
        project_loader("binary-options")

        features.register(DockerFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = actions.get('docker:emit-docker-compose-config')  # type:EmitDockerComposeConfigAction
        action.execute()

        assert len(list(binaries.all())) == 3
        assert binaries.has("npm-simple")
        assert binaries.has("npm-conditions")
        assert binaries.has("mysql")

        npm_simple_set = binaries.get("npm-simple")
        assert len(npm_simple_set) == 1
        npm_simple = list(npm_simple_set)[0]
        assert npm_simple.command() == (''.join(
            effective_command(
                "docker-compose")) + " run --rm --workdir=/app/. --label traefik.enable=false node").split()
        assert npm_simple.command("serve") == (''.join(
            effective_command(
                "docker-compose")) + ' run --rm --workdir=/app/. --label traefik.enable=false node').split()
        assert npm_simple.command("run serve") == (''.join(
            effective_command(
                "docker-compose")) + ' run --rm --workdir=/app/. --label traefik.enable=false node').split()

        npm_conditions_set = binaries.get("npm-conditions")
        assert len(npm_conditions_set) == 1
        npm_conditions = list(npm_conditions_set)[0]
        assert npm_conditions.command() == (''.join(
            effective_command(
                "docker-compose")) + " run --rm --workdir=/app/. --label traefik.enable=false node").split()
        assert npm_conditions.command("serve") == (''.join(
            effective_command(
                "docker-compose")) + ' run --rm --workdir=/app/. --label traefik.enable=false node').split()
        assert npm_conditions.command("run serve") == (
                ''.join(effective_command("docker-compose")) + ' run --rm --workdir=/app/. node').split()

        mysql_set = binaries.get("mysql")
        assert len(mysql_set) == 1
        mysql = list(mysql_set)[0]
        assert mysql.command() == (''.join(effective_command(
            "docker-compose")) + ' run --rm --workdir=/app/. db mysql -hdb -uproject-management-tool -pproject-management-tool').split()

    def test_local_volume_simple(self, project_loader):
        project_loader("local-volume-simple")

        features.register(DockerFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = actions.get('docker:emit-docker-compose-config')  # type:EmitDockerComposeConfigAction
        action.execute()

        assert not os.path.exists('node-path')
        assert os.path.isdir('existing-directory')
        assert os.path.isdir('new_directory')
        assert os.path.isfile('existing-file.txt')
        assert os.path.isfile('new_file')
        assert os.path.isfile('another_new_file.txt')

        with open('existing-file.txt', 'r') as f:
            assert f.read() == 'existing-file.txt'

    def test_local_volume_related(self, project_loader):
        project_loader("local-volume-related")

        features.register(DockerFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = actions.get('docker:emit-docker-compose-config')  # type:EmitDockerComposeConfigAction
        action.execute()

        assert not os.path.exists('node-path')
        assert os.path.isdir('new_directory')
        assert os.path.isdir('child')
        assert os.path.isdir(os.path.join('new_directory', 'some', 'child'))

    @pytest.mark.docker
    def test_traefik_cert(self, project_loader, module_scoped_container_getter):
        project_loader("traefik-cert")

        features.register(CertsFeature())
        features.register(TraefikFeature())
        features.register(DockerFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        setup_cfssl(module_scoped_container_getter)

        shutil.copyfile("docker-compose.yml", "docker-compose.original.yml")

        action = actions.get('docker:emit-docker-compose-config')  # type:EmitDockerComposeConfigAction
        action.execute()

        assert os.path.exists(os.path.join(config.paths.project_home, ".certs", "web.domain.tld.key"))
        assert os.path.exists(os.path.join(config.paths.project_home, ".certs", "web.domain.tld.crt"))
        assert os.path.exists(os.path.join(config.paths.home, "certs", "web.domain.tld.key"))
        assert os.path.exists(os.path.join(config.paths.home, "certs", "web.domain.tld.crt"))
        assert os.path.exists(os.path.join(config.paths.home, "traefik", "config", "web.domain.tld.ssl.toml"))

        shutil.copyfile("docker-compose.removed.yml", "docker-compose.yml")
        action.execute()

        assert not os.path.exists(os.path.join(config.paths.project_home, ".certs", "web.domain.tld.key"))
        assert not os.path.exists(os.path.join(config.paths.project_home, ".certs", "web.domain.tld.crt"))
        assert not os.path.exists(os.path.join(config.paths.home, "certs", "web.domain.tld.key"))
        assert not os.path.exists(os.path.join(config.paths.home, "certs", "web.domain.tld.crt"))
        assert not os.path.exists(os.path.join(config.paths.home, "traefik", "config", "web.domain.tld.ssl.toml"))

        shutil.copyfile("docker-compose.original.yml", "docker-compose.yml")
        action.execute()

        assert os.path.exists(os.path.join(config.paths.project_home, ".certs", "web.domain.tld.key"))
        assert os.path.exists(os.path.join(config.paths.project_home, ".certs", "web.domain.tld.crt"))
        assert os.path.exists(os.path.join(config.paths.home, "certs", "web.domain.tld.key"))
        assert os.path.exists(os.path.join(config.paths.home, "certs", "web.domain.tld.crt"))
        assert os.path.exists(os.path.join(config.paths.home, "traefik", "config", "web.domain.tld.ssl.toml"))

        shutil.copyfile("docker-compose.changed.yml", "docker-compose.yml")
        action.execute()

        assert not os.path.exists(os.path.join(config.paths.project_home, ".certs", "web.domain.tld.key"))
        assert not os.path.exists(os.path.join(config.paths.project_home, ".certs", "web.domain.tld.crt"))
        assert not os.path.exists(os.path.join(config.paths.home, "certs", "web.domain.tld.key"))
        assert not os.path.exists(os.path.join(config.paths.home, "certs", "web.domain.tld.crt"))
        assert not os.path.exists(os.path.join(config.paths.home, "traefik", "config", "web.domain.tld.ssl.toml"))

        assert os.path.exists(os.path.join(config.paths.project_home, ".certs", "web-changed.domain.tld.key"))
        assert os.path.exists(os.path.join(config.paths.project_home, ".certs", "web-changed.domain.tld.crt"))
        assert os.path.exists(os.path.join(config.paths.home, "certs", "web-changed.domain.tld.key"))
        assert os.path.exists(os.path.join(config.paths.home, "certs", "web-changed.domain.tld.crt"))
        assert os.path.exists(os.path.join(config.paths.home, "traefik", "config", "web-changed.domain.tld.ssl.toml"))

class TestDockerDisplayInfoAction:

    def test_retrieve_environment_data(self):
        features.register(DockerFeature())
        load_registered_features()
        action = actions.get('docker:display-info')  # type:DockerDisplayInfoAction

        assert {} == action._retrieve_environment_data(Dotty({}))
        assert {} == action._retrieve_environment_data(Dotty({'toto': 'toto'}))
        assert {'AZERTY': '123'} == action._retrieve_environment_data(Dotty({'environment': {'AZERTY': '123'}}))

    def test_retrieve_ports_data(self):
        features.register(DockerFeature())
        load_registered_features()
        action = actions.get('docker:display-info')  # type:DockerDisplayInfoAction

        assert [] == action._retrieve_service_ports(Dotty({}))
        assert [] == action._retrieve_service_ports(Dotty({'toto': 'toto'}))
        assert [ServicePort(45, 123, None, None, None)] == action._retrieve_service_ports(
            Dotty({'ports': [{'published': '123', 'target': '45'}]}))
        assert [ServicePort(45, 123, None, None, None)] == action._retrieve_service_ports(Dotty({'ports': ['123:45']}))
        assert [ServicePort(45, 123, 'tcp', None, None)] == action._retrieve_service_ports(
            Dotty({'ports': ['123:45/tcp']}))

    def test_retrieve_binaries_data(self):
        features.register(DockerFeature())
        load_registered_features()
        action = actions.get('docker:display-info')  # type:DockerDisplayInfoAction

        assert [] == action._retrieve_binaries_data(Dotty({}))
        assert [] == action._retrieve_binaries_data(Dotty({'toto': 'toto'}))
        assert [] == action._retrieve_binaries_data(Dotty({'labels': {'toto': '123'}}))
        assert ['npm-simple'] == action._retrieve_binaries_data(Dotty({'labels': {
            'ddb.emit.docker:binary[npm-simple](name)': 'npm-simple',
            'ddb.emit.docker:binary[npm-simple](workdir)': '/app'
        }}))
        assert ['npm', 'npm-simple'] == sorted(action._retrieve_binaries_data(Dotty({'labels': {
            'ddb.emit.docker:binary[npm](name)': 'npm',
            'ddb.emit.docker:binary[npm-simple](name)': 'npm-simple',
            'ddb.emit.docker:binary[npm-simple](workdir)': '/app'
        }})))

    def test_retrieve_vhosts_data(self):
        features.register(DockerFeature())
        load_registered_features()
        action = actions.get('docker:display-info')  # type:DockerDisplayInfoAction

        assert [] == action._retrieve_vhosts_data(Dotty({}))
        assert [] == action._retrieve_vhosts_data(Dotty({'toto': 'toto'}))
        assert [] == action._retrieve_vhosts_data(Dotty({'labels': {'toto': '123'}}))
        assert ['http://web.domain.tld/'] == action._retrieve_vhosts_data(Dotty({'labels': {
            'traefik.http.routers.defaults-project.rule': 'Host(`web.domain.tld`)',
            'ddb.emit.docker:binary[npm-simple](workdir)': '/app'
        }}))
        assert ['http://test.tld/', 'http://web.domain.tld/'] == sorted(action._retrieve_vhosts_data(Dotty({'labels': {
            'traefik.http.routers.defaults-project.rule': 'Host(`web.domain.tld`)',
            'traefik.http.routers.test-project.rule': 'Host(`test.tld`)',
            'ddb.emit.docker:binary[npm-simple](workdir)': '/app'
        }})))
        assert ['http://test.tld/', 'https://web.domain.tld/'] == sorted(action._retrieve_vhosts_data(Dotty({'labels': {
            'traefik.http.routers.defaults-project.rule': 'Host(`web.domain.tld`)',
            'traefik.http.routers.defaults-project-tls.rule': 'Host(`web.domain.tld`)',
            'traefik.http.routers.test-project.rule': 'Host(`test.tld`)',
            'ddb.emit.docker:binary[npm-simple](workdir)': '/app'
        }})))

    def test_output_data(self, project_loader):
        features.register(DockerFeature())
        load_registered_features()
        action = actions.get('docker:display-info')  # type:DockerDisplayInfoAction

        config.args.type = None

        assert '' == action._output_data('test', dict(), [], [], [])

        test_vhosts = action._retrieve_vhosts_data(Dotty({'labels': {
            'traefik.http.routers.defaults-project.rule': 'Host(`web.domain.tld`)',
            'traefik.http.routers.defaults-project-tls.rule': 'Host(`web.domain.tld`)',
            'traefik.http.routers.test-project.rule': 'Host(`test.tld`)',
            'ddb.emit.docker:binary[npm-simple](workdir)': '/app'
        }}))

        assert ('\n'.join(['+-------------------------+',
                           '| test                    |',
                           '+-------------------------+',
                           '| http://test.tld/        |',
                           '| https://web.domain.tld/ |',
                           '+-------------------------+'])) == action._output_data('test', dict(), [],
                                                                                   [], test_vhosts)

        test_binaries = action._retrieve_binaries_data(Dotty({'labels': {
            'ddb.emit.docker:binary[npm](name)': 'npm',
            'ddb.emit.docker:binary[npm-simple](name)': 'npm-simple',
            'ddb.emit.docker:binary[npm-simple](workdir)': '/app'
        }}))

        assert ('\n'.join(['+-------------------------+',
                           '| test                    |',
                           '+-------------------------+',
                           '| npm                     |',
                           '| npm-simple              |',
                           '+-------------------------+',
                           '| http://test.tld/        |',
                           '| https://web.domain.tld/ |',
                           '+-------------------------+'])) == action._output_data('test', dict(), [],
                                                                                   test_binaries, test_vhosts)

        test_envs = action._retrieve_environment_data(Dotty({'environment': {'AZERTY': '123'}}))
        assert ('\n'.join(['+-------------------------+',
                           '| test                    |',
                           '+-------------------------+',
                           '| AZERTY: 123             |',
                           '+-------------------------+',
                           '| npm                     |',
                           '| npm-simple              |',
                           '+-------------------------+',
                           '| http://test.tld/        |',
                           '| https://web.domain.tld/ |',
                           '+-------------------------+'])) == action._output_data('test', test_envs, [],
                                                                                   test_binaries, test_vhosts)

        test_ports = action._retrieve_service_ports(Dotty({'ports': [{'published': '12123', 'target': '123'}]}))
        assert ('\n'.join(['+-------------------------+',
                           '| test                    |',
                           '+-------------------------+',
                           '| AZERTY: 123             |',
                           '+-------------------------+',
                           '| 12123:123/tcp           |',
                           '+-------------------------+',
                           '| npm                     |',
                           '| npm-simple              |',
                           '+-------------------------+',
                           '| http://test.tld/        |',
                           '| https://web.domain.tld/ |',
                           '+-------------------------+'])) == action._output_data('test', test_envs, test_ports,
                                                                                   test_binaries, test_vhosts)

        config.args.type = 'nothing'
        assert '' == action._output_data('test', dict(), [], [], test_vhosts)

        config.args.type = 'vhost'
        assert ('\n'.join(['+-------------------------+',
                           '| test                    |',
                           '+-------------------------+',
                           '| http://test.tld/        |',
                           '| https://web.domain.tld/ |',
                           '+-------------------------+'])) == action._output_data('test', dict(), [],
                                                                                   [], test_vhosts)

    def test_execute_traefik_cert(self, capsys: CaptureFixture, project_loader):
        project_loader("traefik-cert")

        features.register(DockerFeature())
        load_registered_features()
        config.args.type = None
        action = actions.get('docker:display-info')  # type:DockerDisplayInfoAction
        action.execute()

        capture = capsys.readouterr()
        assert capture.out
        assert not capture.err
        assert ('\n'.join(['+-------------------------+',
                           '| web                     |',
                           '+-------------------------+',
                           '| https://web.domain.tld/ |',
                           '+-------------------------+',
                           '\n'])) == capture.out

    def test_execute_extra_services(self, capsys: CaptureFixture, project_loader):
        project_loader("extra-services")

        features.register(DockerFeature())
        features.register(TraefikFeature())
        load_registered_features()
        config.args.type = None
        action = actions.get('docker:display-info')  # type:DockerDisplayInfoAction
        action.execute()

        capture = capsys.readouterr()
        assert capture.out
        assert not capture.err
        assert ('\n'.join(['+-------------------------+',
                           '| web                     |',
                           '+-------------------------+',
                           '| https://web.domain.tld/ |',
                           '+-------------------------+',
                           '',
                           '+-----------------------------------------------+',
                           '| foo (extra)                                   |',
                           '+-----------------------------------------------+',
                           '| https://sub.test --> http://192.168.99.1:8080 |',
                           '+-----------------------------------------------+',
                           '\n'])) == capture.out


class TestDockerBinary:
    def test_docker_binary_compare_ne(self):
        binary1 = DockerBinary("npm", "node1")
        binary2 = DockerBinary("npm", "node2")
        binary3 = DockerBinary("npm", "node3")
        binary4 = DockerBinary("npm", "node4")

        assert binary1 != binary2 != binary3 != binary4
        assert hash(binary1) != hash(binary2) != hash(binary3) != hash(binary4)

    def test_docker_binary_compare_eq(self):
        binary1 = DockerBinary("npm", "node")
        binary2 = DockerBinary("npm", "node")

        assert binary1 == binary2
        assert hash(binary1) == hash(binary2)

    def test_docker_binary_sort(self):
        binary1 = DockerBinary("npm", "node1")
        binary2 = DockerBinary("npm", "node2")
        binary3 = DefaultBinary("npm", ["npm"])
        binary4 = DockerBinary("npm", "node4")

        bins = (binary1, binary2, binary3, binary4)
        sorted_bins = tuple(sorted(bins))
        expected = (binary1, binary2, binary4, binary3)

        assert sorted_bins == expected

        binary_with_condition = DockerBinary("npm", "node5", condition="some-condition")

        bins = (binary1, binary2, binary_with_condition, binary3, binary4)
        sorted_bins = tuple(sorted(bins))
        expected = (binary_with_condition, binary1, binary2, binary4, binary3)

        assert sorted_bins == expected

    def test_docker_utils_get_mapped_path(self):
        assert get_mapped_path('/home/toilal/projects/test') == '/home/toilal/projects/test'

        config.data['docker.path_mapping'] = {
            '/home/toilal/projects': '/projects',
            '/home/toilal': '/home'
        }

        assert get_mapped_path('/home/toilal/projects/test') == '/projects/test'



# -*- coding: utf-8 -*-
import os
import pathlib
from typing import Optional, List

import zgitignore

from ddb.action.action import EventBinding, InitializableAction
from ddb.cache import caches, register_project_cache
from ddb.config import config
from ddb.context import context
from ddb.event import events
from ddb.utils.file import force_remove


class UpdateGitignoreAction(InitializableAction):
    """
    Append generated files to .gitignore
    """

    @property
    def name(self) -> str:
        return "git:update-gitignore"

    @property
    def event_bindings(self):
        return (events.file.generated,
                EventBinding(events.file.deleted, self.remove),
                EventBinding(events.phase.configure, self.enforced)
                )

    def initialize(self):
        register_project_cache("gitignore")

    def destroy(self):
        if caches.has("gitignore"):
            caches.unregister("gitignore", callback=lambda c: c.close())

    @staticmethod
    def execute(target: str, source: str = None):
        """
        Execute action
        """
        if not UpdateGitignoreAction._check_file_in_project(target):
            return

        ignored, _, _, _, _, _ = UpdateGitignoreAction._check_file_ignored(target)
        if ignored:
            return

        try:
            gitignore = next(UpdateGitignoreAction._find_gitignores(target))
        except StopIteration:
            gitignore = ".gitignore"

        relative_target = UpdateGitignoreAction._get_relative_path(target=target, gitignore=gitignore, first_slash=True)

        # Retrieve the block lines
        block_lines, before_lines, after_lines = UpdateGitignoreAction._read_and_split_gitignore_file(gitignore)

        # Add of the relative_target to the block and sort the whole block.
        if relative_target not in block_lines:
            block_lines.append(relative_target)

            UpdateGitignoreAction._sanitize_block(block_lines, before_lines, after_lines)
            UpdateGitignoreAction._write_gitignore_content(gitignore, before_lines + block_lines + after_lines)
            context.log.success("%s added to %s", relative_target, gitignore)

    @staticmethod
    def enforced():
        """
        Handle the list of enforced actions
        """
        files = config.data.get('gitignore.enforce')
        cache = caches.get("gitignore")
        cached_files = cache.get('enforced', [])

        for file in cached_files:
            if file not in files:
                UpdateGitignoreAction.remove(file)

        for file in files:
            UpdateGitignoreAction.execute(file)
            if file not in cached_files:
                cached_files.append(file)

        cache.set('enforced', cached_files)
        cache.flush()

    @staticmethod
    def remove(file: str):
        """
        Remove an ignored file
        :param file:
        :return:
        """
        ignored, target_file, gitignore, block_lines, before_lines, after_lines = \
            UpdateGitignoreAction._check_file_ignored(file)
        if not ignored:
            return

        # Removal of the target_file to the block and sort the whole block.
        if target_file in block_lines:
            block_lines.remove(target_file)

            UpdateGitignoreAction._sanitize_block(block_lines, before_lines, after_lines)
            UpdateGitignoreAction._write_gitignore_content(gitignore, before_lines + block_lines + after_lines)

            context.log.warning("%s removed from %s", file, gitignore)

    @staticmethod
    def _check_file_in_project(target):
        target_path = os.path.realpath(target)
        cwd = os.path.realpath(".")
        return target_path.startswith(cwd)

    @staticmethod
    def _get_relative_path(target: str, gitignore: str, first_slash: bool = True):
        relpath = pathlib.Path(os.path.relpath(target, os.path.dirname(gitignore))).as_posix()
        if first_slash:
            return "/" + relpath
        return relpath

    @staticmethod
    def _find_gitignores(target: str):
        """
        Find writable .gitignore files, walking directories up to project home.
        """
        dirname = target

        while os.path.abspath(dirname).startswith(config.paths.project_home) and \
                os.path.abspath(dirname) != os.path.abspath(config.paths.project_home):
            dirname = os.path.dirname(dirname)
            gitignore = os.path.join(dirname, ".gitignore")
            if os.path.exists(gitignore) and os.access(gitignore, os.W_OK):
                yield gitignore

    @staticmethod
    def _check_file_ignored(file: str, gitignores=None):
        """
        Check if a file is ignored, going throw all provided gitignore files
        :param file:
        :param gitignores:
        :return:
        """
        if not gitignores:
            gitignores = UpdateGitignoreAction._find_gitignores(file)

        for gitignore in gitignores:
            block_lines, before_lines, after_lines = UpdateGitignoreAction._read_and_split_gitignore_file(gitignore)

            gitignore_content = before_lines + block_lines + after_lines

            inversed_gitignore_content = []
            for pattern in gitignore_content:
                if pattern.startswith("!"):
                    inversed_gitignore_content.append(pattern[1:])

            relative_file = UpdateGitignoreAction._get_relative_path(file, gitignore)

            zgitignore_helper = zgitignore.ZgitIgnore(gitignore_content)
            inversed_zgitignore_helper = zgitignore.ZgitIgnore(inversed_gitignore_content)
            if zgitignore_helper.is_ignored(relative_file) or inversed_zgitignore_helper.is_ignored(relative_file):
                return True, relative_file, gitignore, block_lines, before_lines, after_lines
        return False, None, None, None, None, None

    @staticmethod
    def _get_block_limit(start: bool):
        """
        Generate the gitignore block limit
        :param start: if it is the start of the block or not
        :return:
        """
        markers = config.data.get('gitignore.markers', ['inetum-orleans/docker-devbox'])

        if start:
            return f'###> {markers[0]} ###'
        return f'###< {markers[0]} ###'

    @staticmethod
    def _is_block_limit(line: str, start: Optional[bool] = None):
        """
        Check if line match a block limit.
        :param line: line to check
        :param start: if it is the start of the block or not
        :return:
        """
        markers = config.data.get('gitignore.markers', ['inetum-orleans/docker-devbox'])

        for marker in markers:
            if start or start is None:
                if f'###> {marker} ###' == line:
                    return True
            if not start or start is None:
                if f'###< {marker} ###' == line:
                    return True
        return False

    @staticmethod
    def _write_gitignore_content(gitignore: str, content: List[str]):
        if [line for line in content if line]:
            with open(gitignore, "w", encoding="utf-8") as gitignore_file:
                for lineno, line in enumerate(content):
                    if lineno > 0:
                        gitignore_file.write('\n')
                    gitignore_file.write(line)
        else:
            force_remove(gitignore)

    @staticmethod
    def _split_block_lines(gitignore_file_content: str):
        started = False
        ended = False
        before_lines = []
        content_lines = []
        after_lines = []
        for line in gitignore_file_content.split('\n'):
            if UpdateGitignoreAction._is_block_limit(line, False):
                ended = True
            if not started and not ended:
                before_lines.append(line)
            elif started and not ended:
                content_lines.append(line)
            elif started and ended:
                after_lines.append(line)
            if UpdateGitignoreAction._is_block_limit(line, True):
                started = True

        return content_lines, before_lines, after_lines

    @staticmethod
    def _read_and_split_gitignore_file(gitignore_file: str):
        if os.path.exists(gitignore_file):
            with open(gitignore_file, "r+", encoding="utf-8") as gitignore_fp:
                return UpdateGitignoreAction._split_block_lines(gitignore_fp.read())
        else:
            return [], [], []

    @staticmethod
    def _sanitize_block(block_lines, before_lines, after_lines):
        block_lines = [line for line in block_lines if line]
        block_lines.sort()

        if block_lines:
            if not before_lines or not UpdateGitignoreAction._is_block_limit(before_lines[-1], start=True):
                before_lines.append(UpdateGitignoreAction._get_block_limit(start=True))

            if not after_lines or not UpdateGitignoreAction._is_block_limit(after_lines[0], start=False):
                after_lines.insert(0, UpdateGitignoreAction._get_block_limit(start=False))
        else:
            if before_lines and UpdateGitignoreAction._is_block_limit(before_lines[-1], start=True):
                before_lines.pop(-1)
            if after_lines and UpdateGitignoreAction._is_block_limit(after_lines[0], start=False):
                after_lines.pop(0)



from .action import Action, InitializableAction
from .runner import InitializableActionEventBindingRunner, ActionEventBindingRunner, EventBindingRunner


def action_event_binding_runner_factory(action: Action, event_name: str, to_call=None, event_processor=None,
                                        fail_fast=False) -> EventBindingRunner:
    """
    Create an event binding runner from an action, event_name and optional callable.
    """
    if isinstance(action, InitializableAction):
        return InitializableActionEventBindingRunner(action, event_name, to_call, event_processor, fail_fast=fail_fast)
    return ActionEventBindingRunner(action, event_name, to_call, event_processor, fail_fast=fail_fast)



import os

from ddb.__main__ import main
import zipfile


class TestDjp:
    def test_djp(self, project_loader):
        project_loader("djp")

        with zipfile.ZipFile('template.zip', 'r') as zip_ref:
            zip_ref.extractall('expected')

        main(["download"])

        assert os.path.exists('.docker/djp/Dockerfile.jinja')
        assert os.path.exists('.docker/djp/djp.libjsonnet')

        with open('.docker/djp/djp.libjsonnet') as djp_libjsonnet:
            assert not '// Edited\n' in djp_libjsonnet.readlines()

    def test_djp_patch(self, project_loader):
        project_loader("djp_patch")

        main(["download"])

        assert os.path.exists('.docker/djp/Dockerfile.jinja')
        assert os.path.exists('.docker/djp/djp.libjsonnet')

        with open('.docker/djp/Dockerfile.jinja') as djp_libjsonnet:
            assert 'replaced' not in djp_libjsonnet.read()

        with open('.docker/djp/djp.libjsonnet') as djp_libjsonnet:
            assert '// Edited\n' in djp_libjsonnet.readlines()



# -*- coding: utf-8 -*-
from argparse import ArgumentParser
from typing import Iterable, ClassVar

from .actions import RunAction
from .schema import RunSchema
from ..feature import Feature
from ..schema import FeatureSchema
from ...action import Action
from ...command import LifecycleCommand, Command
from ...command.command import LifecycleWithoutHelpCommand
from ...phase import Phase, DefaultPhase


class RunFeature(Feature):
    """
    Run registered binaries.
    """

    @property
    def name(self) -> str:
        return "binary"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core"]

    @property
    def schema(self) -> ClassVar[FeatureSchema]:
        return RunSchema

    @property
    def phases(self) -> Iterable[Phase]:
        def run_parser(parser: ArgumentParser):
            parser.add_argument("name", help="Binary name to run")

        return (
            DefaultPhase("run", "Display command to run project binary",
                         run_parser,
                         allow_unknown_args=True),
        )

    @property
    def commands(self) -> Iterable[Command]:
        return (
            LifecycleWithoutHelpCommand("run", "Display command to run project binary",
                                        "run", avoid_stdout=True),
        )

    @property
    def actions(self) -> Iterable[Action]:
        return (
            RunAction(),
        )



# -*- coding: utf-8 -*-
import os

from marshmallow import fields, Schema

from ddb.feature.schema import FeatureSchema


class PathSchema(Schema):
    """
    Path settings for core feature schema.
    """
    project_home = fields.String(default=None)  # default is set in feature _configure_defaults
    home = fields.String(default=None, allow_none=True)  # default is set in feature _configure_defaults
    ddb_home = fields.String(default=None, allow_none=True)  # default is set in feature _configure_defaults


class EnvSchema(Schema):
    """
    Env settings for core feature schema.
    """
    current = fields.String(required=True, default=None)  # default is set in feature _configure_defaults
    available = fields.List(fields.String, required=True, default=["prod", "stage", "ci", "dev"])


class DomainSchema(Schema):
    """
    Domain settings for core feature schema.
    """
    sub = fields.String(required=True, default=None)  # default is set in feature _configure_defaults
    ext = fields.String(required=True, default="test")
    value = fields.String(required=True, default=None)  # default is set in feature _configure_defaults


class ProjectSchema(Schema):
    """
    Env settings for core feature schema.
    """
    name = fields.String(required=True, default=None)  # default is set in feature _configure_defaults


class ProcessSchema(Schema):
    """
    Process settings for core feature schema.
    """
    bin = fields.String(required=False, allow_none=True, default=None)
    prepend = fields.Raw(required=False, allow_none=True, default=None)  # List[str] or str splitted with shlex
    append = fields.Raw(required=False, allow_none=True, default=None)  # List[str] or str splitted with shlex


class ConfigurationSchema(Schema):
    """
    Process settings for core feature schema.
    """
    extra = fields.List(fields.String(), required=False)


class CoreFeatureSchema(FeatureSchema):
    """
    Core feature schema.
    """
    env = fields.Nested(EnvSchema(), default=EnvSchema())
    domain = fields.Nested(DomainSchema(), default=DomainSchema())
    project = fields.Nested(ProjectSchema(), default=ProjectSchema())
    os = fields.String(required=True, default=os.name)
    path = fields.Nested(PathSchema(), default=PathSchema())
    process = fields.Dict(fields.String(), fields.Nested(ProcessSchema()), default={})  # Process binary mappings
    configuration = fields.Nested(ConfigurationSchema(), default=ConfigurationSchema())
    github_repository = fields.String(required=True, default="inetum-orleans/docker-devbox-ddb")
    check_updates = fields.Boolean(required=True, default=True)
    required_version = fields.String(required=False, allow_none=True, default=None)
    release_asset_name = fields.String(required=False, allow_none=True,
                                       default=None)  # default is set in feature _configure_defaults



import os
import yaml

from ddb.utils.process import run


class DockerUtils:
    """
    A set of tools to manipulate docker using docker and docker-compose system commands
    """

    @staticmethod
    def service_up(name: str = None):
        """
        Execute docker-compose up -d
        :param name: the name of a specific service to up
        :raise DockerComposeYamlMissingException: in case of missing docker-compose.yml in folder
        :return:
        """
        if not os.path.exists("docker-compose.yml"):
            raise DockerComposeYamlMissingException
        if name is None:
            run("docker-compose", "up", "-d")
        else:
            run("docker-compose", "up", "-d", name)

    @staticmethod
    def service_stop(name: str):
        """
        Execute docker-compose stop
        :param name: the name of a specific service to stop
        :raise DockerComposeYamlMissingException: in case of missing docker-compose.yml in folder
        :return:
        """
        if not os.path.exists("docker-compose.yml"):
            raise DockerComposeYamlMissingException
        if name is None:
            run("docker-compose", "stop")
        else:
            run("docker-compose", "stop", name)

    @staticmethod
    def is_container_up(name: str):
        """
        Check if the given service is up.
        :param name: The name of the container to check
        :raise DockerComposeYamlMissingException: in case of missing docker-compose.yml in folder
        :return:
        """
        if not os.path.exists("docker-compose.yml"):
            raise DockerComposeYamlMissingException
        container_id = run("docker-compose", "ps", "-q", name).decode("utf-8").rstrip()
        if len(container_id) == 0:
            return False
        containers = run('docker', 'ps', '-q', '--no-trunc').decode("utf-8").rstrip().split('\n')
        return container_id in containers

    @staticmethod
    def get_config():
        """
        Get the docker-compose.yml configuration content.
        :raise DockerComposeYamlMissingException: in case of missing docker-compose.yml
        :return:
        """
        if not os.path.exists("docker-compose.yml"):
            raise DockerComposeYamlMissingException

        config = run("docker-compose", "config")
        return yaml.load(config, yaml.SafeLoader)


class DockerComposeYamlMissingException(Exception):
    """
    Exception raised in case of missing docker-compose file.
    """

    def __init__(self):
        Exception.__init__()
        self.message = "There is no docker-compose.yml file in current folder ({})".format(os.getcwd())



from ddb.feature.core.actions import is_version_greater


class TestCoreFeature:
    def test_is_version_greater(self):
        assert is_version_greater('1.9.2', '1.10.0')
        assert not is_version_greater('1.10.0', '1.9.2')
        assert not is_version_greater('1.10.0', '1.10.0')

        assert is_version_greater('v1.9.2', '1.10.0')
        assert not is_version_greater('v1.10.0', '1.9.2')
        assert not is_version_greater('v1.10.0', '1.10.0')

        assert is_version_greater('1.9.2', 'v1.10.0')
        assert not is_version_greater('1.10.0', 'v1.9.2')
        assert not is_version_greater('1.10.0', 'v1.10.0')

        assert is_version_greater('v1.9.2', 'v1.10.0')
        assert not is_version_greater('v1.10.0', 'v1.9.2')
        assert not is_version_greater('v1.10.0', 'v1.10.0')



import os

from ddb.__main__ import load_registered_features
from ddb.feature import features
from ddb.feature.cookiecutter import CookiecutterFeature
from ddb.feature.cookiecutter.actions import CookiecutterAction


class TestCookiecutterAction:
    def test_empty_project(self, project_loader):
        project_loader("empty")

        features.register(CookiecutterFeature())
        load_registered_features()

        action = CookiecutterAction()
        action.execute()

    def test_github_django(self, project_loader):
        project_loader("github-django")

        features.register(CookiecutterFeature())
        load_registered_features()

        action = CookiecutterAction()
        action.execute()

        assert os.path.isdir('my_awesome_project')
        assert os.path.isfile(os.path.join('my_awesome_project', 'setup.cfg'))



from collections import namedtuple

from .errors import ConfigurationError
from docker.utils.ports import build_port_bindings


class ServicePort(namedtuple('_ServicePort', 'target published protocol mode external_ip')):
    def __new__(cls, target, published, *args, **kwargs):
        try:
            if target:
                target = int(target)
        except ValueError:
            raise ConfigurationError('Invalid target port: {}'.format(target))

        if published:
            if isinstance(published, str) and '-' in published:  # "x-y:z" format
                a, b = published.split('-', 1)
                if not a.isdigit() or not b.isdigit():
                    raise ConfigurationError('Invalid published port: {}'.format(published))
            else:
                try:
                    published = int(published)
                except ValueError:
                    raise ConfigurationError('Invalid published port: {}'.format(published))

        return super().__new__(
            cls, target, published, *args, **kwargs
        )

    @classmethod
    def parse(cls, spec):
        if isinstance(spec, cls):
            # When extending a service with ports, the port definitions have already been parsed
            return [spec]

        if not isinstance(spec, dict):
            result = []
            try:
                for k, v in build_port_bindings([spec]).items():
                    if '/' in k:
                        target, proto = k.split('/', 1)
                    else:
                        target, proto = (k, None)
                    for pub in v:
                        if pub is None:
                            result.append(
                                cls(target, None, proto, None, None)
                            )
                        elif isinstance(pub, tuple):
                            result.append(
                                cls(target, pub[1], proto, None, pub[0])
                            )
                        else:
                            result.append(
                                cls(target, pub, proto, None, None)
                            )
            except ValueError as e:
                raise ConfigurationError(str(e))

            return result

        return [cls(
            spec.get('target'),
            spec.get('published'),
            spec.get('protocol'),
            spec.get('mode'),
            None
        )]

    @property
    def merge_field(self):
        return (self.target, self.published, self.external_ip, self.protocol)

    def repr(self):
        return {
            k: v for k, v in zip(self._fields, self) if v is not None
        }

    def legacy_repr(self):
        return normalize_port_dict(self.repr())


def normalize_port_dict(port):
    return '{external_ip}{has_ext_ip}{published}{is_pub}{target}/{protocol}'.format(
        published=port.get('published', ''),
        is_pub=(':' if port.get('published') is not None or port.get('external_ip') else ''),
        target=port.get('target'),
        protocol=port.get('protocol', 'tcp'),
        external_ip=port.get('external_ip', ''),
        has_ext_ip=(':' if port.get('external_ip') else ''),
    )



# -*- coding: utf-8 -*-
from subprocess import CalledProcessError
from typing import ClassVar, Iterable

from dotty_dict import Dotty

from ddb.feature import Feature
from .schema import VersionSchema
from ...utils.process import run


def is_git_repository():
    """
    Check if git command line is available, and if a git repository index is available in current directory
    """
    try:
        run("git", "rev-parse", "--git-dir")
        return True
    except CalledProcessError:
        return False


def get_tag_from_vcs():
    """
    Get last tag from git index
    """
    try:
        return run("git", "describe", "--tags", "--abbrev=0").decode("utf-8").strip()
    except CalledProcessError as exc:
        if exc.returncode == 128:
            return None
        raise


def get_branch_from_vcs():
    """
    Get branch name from git index
    """
    try:
        branch = run("git", "rev-parse", "--abbrev-ref", "HEAD").decode("utf-8").strip()
    except CalledProcessError as exc:
        if exc.returncode == 128:
            return None
        raise

    if branch == "HEAD":
        # CI checkout a commit hash, making the HEAD detached.
        try:
            refs = run("git", "for-each-ref", "--format=%(refname:short)", "refs/heads") \
                .decode("utf-8").strip()

            if refs:
                refs_split = refs.splitlines()
                if refs_split:
                    return refs_split[0]
        except CalledProcessError as exc:
            if exc.returncode == 128:
                return branch
            raise

        # CI may not checkout branches locally, so try with remote origin
        try:
            remote = 'origin'
            refs = run("git", "for-each-ref", "--format=%(refname:short)", "refs/remotes/" + remote) \
                .decode("utf-8").strip()

            if refs:
                refs_split = refs.splitlines()
                if refs_split:
                    value = refs_split[0]
                    if value.startswith(remote + '/'):
                        return value[len(remote + 'origin/'):]
        except CalledProcessError as exc:
            if exc.returncode == 128:
                return branch
            raise

    return branch


def get_version_from_vcs():
    """
    Get version from git index
    """
    try:
        return run("git", "describe", "--tags").decode("utf-8").strip()
    except CalledProcessError as exc:
        if exc.returncode == 128:
            return None
        raise


def get_hash_from_vcs():
    """
    Get commit hash from git index
    """
    try:
        return run("git", "rev-parse", "HEAD").decode("utf-8").strip()
    except CalledProcessError as exc:
        if exc.returncode == 128:
            return None
        raise


def get_short_hash_from_vcs():
    """
    Get commit short hash from git index
    """
    try:
        return run("git", "log", "--pretty=format:%h", "-n", "1").decode("utf-8").strip()
    except CalledProcessError as exc:
        if exc.returncode == 128:
            return None
        raise


class VersionFeature(Feature):
    """
    Update gitignore files when a file is generated.
    """

    @property
    def name(self) -> str:
        return "version"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core"]

    @property
    def schema(self) -> ClassVar[VersionSchema]:
        return VersionSchema

    def _configure_defaults(self, feature_config: Dotty):
        if not is_git_repository():
            return

        branch = feature_config.get("branch")
        if branch is None:
            branch = get_branch_from_vcs()
            feature_config["branch"] = branch

        version = feature_config.get("version")
        if version is None:
            version = get_version_from_vcs()
            feature_config["version"] = version

        tag = feature_config.get("tag")
        if tag is None:
            tag = get_tag_from_vcs()
            feature_config["tag"] = tag

        hash_value = feature_config.get("hash")
        if hash_value is None:
            hash_value = get_hash_from_vcs()
            feature_config["hash"] = hash_value

        short_hash = feature_config.get("short_hash")
        if short_hash is None:
            short_hash = get_short_hash_from_vcs()
            feature_config["short_hash"] = short_hash



# -*- coding: utf-8 -*-
from typing import Iterable, ClassVar

from .actions import GenerateCertAction, RemoveCertAction, GenerateSignerCertificateAction
from .schema import CertsSchema
from ..feature import Feature
from ..schema import FeatureSchema
from ...action import Action


class CertsFeature(Feature):
    """
    Generate SSL certificates for your project.
    """

    @property
    def name(self) -> str:
        return "certs"

    @property
    def dependencies(self) -> Iterable[str]:
        return ["core"]

    @property
    def schema(self) -> ClassVar[FeatureSchema]:
        return CertsSchema

    @property
    def actions(self) -> Iterable[Action]:
        return (
            GenerateCertAction(),
            RemoveCertAction(),
            GenerateSignerCertificateAction()
        )



# -*- coding: utf-8 -*-
from marshmallow import fields

from ddb.feature.schema import FeatureSchema

keywords = ["and", "elif", "in", "or", "break", "else", "lambda", "pass", "continue", "for",
            "load", "return", "def", "if", "not", "while"]

reserved = ["as", "finally", "nonlocal", "assert", "from", "raise", "class", "global", "try",
            "del", "import", "with", "except", "is", "yield"]


class YttSchema(FeatureSchema):
    """
    Ytt schema.
    """
    suffixes = fields.List(fields.String(), default=[".ytt"])
    includes = fields.List(fields.String())  # default is build automatically from suffixes value
    excludes = fields.List(fields.String())
    extensions = fields.List(fields.String(), default=[".yaml", ".yml", ""])
    depends_suffixes = fields.List(fields.String(), default=[".data", ".overlay"])
    args = fields.List(fields.String(), default=["--ignore-unknown-comments"])
    keywords = fields.List(fields.String(), default=keywords + reserved)
    keywords_escape_format = fields.String(default="%s_")



import pytest
from _pytest.fixtures import FixtureRequest

from ddb.cache.shelve_cache import ShelveCache


class TestShelveAdapter:
    def test_should_raise_value_error_set_after_close(self, request: FixtureRequest):
        adapter = ShelveCache(request.node.name)
        adapter.close()

        with pytest.raises(ValueError):
            adapter.set("foo", "bar")

    def test_should_set_value_properly(self, request: FixtureRequest):
        adapter = ShelveCache(request.node.name)
        adapter.set("foo", "bar")

        adapter.flush()
        assert adapter.get("foo") == "bar"

        adapter.close()

        adapter = ShelveCache(request.node.name)
        assert adapter.get("foo") == "bar"

    def test_should_set_number_value_properly(self, request: FixtureRequest):
        adapter = ShelveCache(request.node.name)
        adapter.set("foo", 1337)

        adapter.flush()
        assert adapter.get("foo") == 1337

        adapter.close()

        adapter = ShelveCache(request.node.name)
        assert adapter.get("foo") == 1337

    def test_should_clear_value_properly(self, request: FixtureRequest):
        adapter = ShelveCache(request.node.name)
        adapter.set("foo", "bar")

        adapter.close()
        adapter = ShelveCache(request.node.name)
        adapter.set("foo", "bar")

        adapter.clear()
        adapter.close()

        adapter = ShelveCache(request.node.name)
        assert adapter.get("foo") is None



import os

import pytest
from ddb.__main__ import load_registered_features, register_actions_in_event_bus
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.feature.feature import FeatureConfigurationError
from ddb.feature.file import FileFeature, FileWalkAction
from ddb.feature.symlinks import SymlinksFeature


class TestSymlinksAction:
    def test_empty_configuration_without_core(self, project_loader):
        project_loader("empty")

        with pytest.raises(FeatureConfigurationError):
            features.register(SymlinksFeature())
            load_registered_features()
            register_actions_in_event_bus(True)

    def test_empty_configuration_with_core(self, project_loader):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(SymlinksFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

    def test_project_1(self, project_loader):
        project_loader("project1")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(SymlinksFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('test')
        assert os.path.islink('test')
        assert os.path.exists('test')

        assert os.path.exists('test.yml')
        assert os.path.islink('test.yml')
        assert os.path.exists('test.yml')

    def test_project_2_fallback(self, project_loader):
        project_loader("project2")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(SymlinksFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.islink('test')
        assert os.readlink('test') == 'test.stage'
        assert os.path.exists('test')

        assert os.path.islink('test.yml')
        assert os.readlink('test.yml') == 'test.dev.yml'
        assert os.path.exists('test.yml')

        assert os.path.islink('test2')
        assert os.readlink('test2') == 'test2.prod'
        assert os.path.exists('test2')

        assert not os.path.islink('test3')

    def test_project_4_subdirectory(self, project_loader):
        project_loader("project4")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(SymlinksFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.islink(os.path.join('subdirectory', 'test.yml'))
        assert os.readlink(os.path.join('subdirectory', 'test.yml')) == 'test.dev.yml'

        assert os.path.exists(os.path.join('subdirectory', 'test.yml'))

        assert not os.path.islink('no.yml')

    def test_project_many_dev(self, project_loader):
        project_loader("project_many_dev")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(SymlinksFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.islink('test.yml')
        assert os.readlink('test.yml') == 'test.dev.yml'

    def test_project_many_prod(self, project_loader):
        project_loader("project_many_prod")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(SymlinksFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.islink('test.yml')
        assert os.readlink('test.yml') == 'test.prod.yml'



import os

from ddb.__main__ import load_registered_features, register_actions_in_event_bus
from ddb.config import config, migrations
from ddb.config.migrations import PropertyMigration
from ddb.feature import features
from ddb.feature.core import CoreFeature
from ddb.feature.file import FileFeature, FileWalkAction
from ddb.feature.jinja import JinjaFeature


class TestJinjaAction:
    def test_empty_project_without_core(self, project_loader):
        project_loader("empty")

        features.register(FileFeature())
        features.register(JinjaFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

    def test_empty_project_with_core(self, project_loader):
        project_loader("empty")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JinjaFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

    def test_project1(self, project_loader):
        project_loader("project1")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JinjaFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('foo.yml')
        with open('foo.yml', 'r') as f:
            foo = f.read()

        assert foo == 'env: dev'

    def test_project2(self, project_loader):
        project_loader("project2")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JinjaFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('foo.yml')
        with open('foo.yml', 'r') as f:
            foo = f.read()

        assert foo == 'env: dev\nincluded: True'

        assert not os.path.exists(os.path.join("partial", "_partial.yml"))
        assert not os.path.exists(os.path.join("partial", "partial.yml"))

    def test_project3(self, project_loader):
        project_loader("project3")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JinjaFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('.foo.yml')
        with open('.foo.yml', 'r') as f:
            foo = f.read()

        assert foo == 'env: dev'

    def test_project4(self, project_loader):
        project_loader("project4")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JinjaFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('foo')
        with open('foo', 'r') as f:
            foo = f.read()

        assert foo == 'env=dev'

    def test_project5(self, project_loader):
        project_loader("project5")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JinjaFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('foo')
        with open('foo', 'r') as f:
            foo = f.read()
        assert foo == 'env=dev'

        assert os.path.exists('foo.yml')
        with open('foo.yml', 'r') as f:
            foo = f.read()

        assert foo == 'env: dev'

    def test_project_keep_trailing_lines_false(self, project_loader):
        project_loader("trailing-newline")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JinjaFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        config.data['jinja.options'] = {"keep_trailing_newline": False}

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('foo.yml')
        with open('foo.yml', 'r') as f:
            foo = f.read()

        assert foo == 'test: trailing'

    def test_project_keep_trailing_lines_default(self, project_loader):
        project_loader("trailing-newline")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JinjaFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('foo.yml')
        with open('foo.yml', 'r') as f:
            foo = f.read()

        assert foo == 'test: trailing\n'

    def test_project_keep_trailing_lines_true(self, project_loader):
        project_loader("trailing-newline")

        features.register(CoreFeature())
        features.register(FileFeature())
        features.register(JinjaFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        config.data['jinja.options'] = {"keep_trailing_newline": True}

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('foo.yml')
        with open('foo.yml', 'r') as f:
            foo = f.read()

        assert foo == 'test: trailing\n'


class TestJinjaAutofix:
    def teardown_method(self, test_method):
        migrations.set_history()

    def test_autofix_variables_only(self, project_loader):
        project_loader("autofix_variables_only")

        config.args.autofix = True

        history = (
            PropertyMigration("old_property",
                              "new_property", since="v1.1.0"),
            PropertyMigration("some.deep.old.property",
                              "some.another.new.property", since="v1.1.0"),
        )

        migrations.set_history(history)

        features.register(FileFeature())
        features.register(JinjaFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        assert os.path.exists('config.properties')
        with open('config.properties', 'r') as f:
            rendered = f.read()

        with open('config.expected.properties', 'r') as f:
            expected = f.read()

        assert rendered == expected

        with open('config.properties.jinja', 'r') as f:
            source = f.read()

        with open('config.properties.autofix', 'r') as f:
            fixed = f.read()

        assert source == fixed

    def test_autofix_eol(self, project_loader):
        project_loader("autofix_eol")

        history = (
            PropertyMigration("docker.compose.network_name",
                              "jsonnet.docker.compose.network_name", since="v1.6.0"),
            PropertyMigration("docker.port_prefix",
                              "jsonnet.docker.expose.port_prefix", since="v1.6.0"),
        )

        migrations.set_history(history)

        features.register(FileFeature())
        features.register(JinjaFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        with open('docker-compose.properties', 'r') as f:
            rendered = f.read()

        with open('docker-compose.expected.properties', 'r') as f:
            expected = f.read()

        assert rendered == expected

    def test_autofix_eol2(self, project_loader):
        project_loader("autofix_eol2")

        config.args.autofix = True

        history = (
            PropertyMigration("docker.compose.network_name",
                              "jsonnet.docker.compose.network_name", since="v1.6.0"),
            PropertyMigration("docker.port_prefix",
                              "jsonnet.docker.expose.port_prefix", since="v1.6.0"),
        )

        migrations.set_history(history)

        features.register(FileFeature())
        features.register(JinjaFeature())
        load_registered_features()
        register_actions_in_event_bus(True)

        action = FileWalkAction()
        action.initialize()
        action.execute()

        with open('msmtprc', 'r') as f:
            rendered = f.read()

        with open('msmtprc.expected', 'r') as f:
            expected = f.read()

        assert rendered == expected

        with open('msmtprc.jinja', 'r') as f:
            template = f.read()

        with open('msmtprc.autofix', 'r') as f:
            autofix = f.read()

        assert autofix == template



# -*- coding: utf-8 -*-
import os
import shelve
import tempfile

from .cache import Cache
from ..config import config

import dbm

# Easy fix for https://github.com/inetum-orleans/docker-devbox-ddb/issues/49
# pylint:disable=protected-access
_dbm_names = dbm._names = ['dbm.gnu', 'dbm.ndbm', 'dbm.dumb']
if 'dbm.dumb' in _dbm_names:
    # Let's make dbm.dumb the preferred implementation
    _dbm_names.remove('dbm.dumb')
    _dbm_names.insert(0, 'dbm.dumb')
    dbm._names = _dbm_names


class ShelveCache(Cache):
    """
    A cache implementation relying of shelve module.
    """

    def __init__(self, namespace: str):
        super().__init__(namespace)

        if config.paths.home:
            path = os.path.join(config.paths.home, "cache")
        else:
            path = os.path.join(tempfile.gettempdir(), "ddb", "cache")
        os.makedirs(path, exist_ok=True)

        self.basename = os.path.join(path, self._namespace)
        if config.clear_cache:
            self._delete_files(self.basename)
        try:
            self._shelf = shelve.open(self.basename)
        except Exception as open_error:  # pylint:disable=broad-except
            if self._delete_files(self.basename):
                try:
                    self._shelf = shelve.open(self.basename)
                except Exception as fallback_error:  # pylint:disable=broad-except
                    raise open_error from fallback_error
            else:
                raise open_error
        if config.clear_cache:
            self._shelf.clear()

    @staticmethod
    def _delete_files(basename):
        deleted = False
        for filename in (basename, f"{basename}.dat", f"{basename}.dir"):
            if os.path.exists(filename):
                os.remove(filename)
                deleted = True
        return deleted

    def close(self):
        self._shelf.close()

    def flush(self):
        self._shelf.sync()

    def get(self, key: str, default=None):
        try:
            return self._shelf.get(key, default)
        except AttributeError:
            # This can occur when class definition hash change.
            self._delete_files(self.basename)
            raise

    def keys(self):
        return self._shelf.keys()

    def set(self, key: str, data):
        self._shelf[key] = data

    def pop(self, key: str):
        try:
            return self._shelf.pop(key)
        except AttributeError:
            # This can occur when class definition hash change.
            self._delete_files(self.basename)
            raise

    def clear(self):
        try:
            self._shelf.clear()
        except AttributeError:
            # This can occur when class definition hash change.
            self._delete_files(self.basename)
            raise

    def __contains__(self, key):
        return self._shelf.__contains__(key)



# -*- coding: utf-8 -*-
import os
from abc import abstractmethod, ABC
from typing import Callable, Iterable, Union, Tuple, Optional

from ddb.cache import register_project_cache, caches
from ddb.config import config
from ddb.context import context
from ddb.event import events
from ddb.registry import RegistryObject
from ddb.utils.file import write_if_different, TemplateFinder, force_remove, copy_if_different


class EventBinding:
    """
    A configured event binding on action.

    event is the event name

    call is a callable that will be invoked with event *args, **kwargs.

    processor is used to process the event before invoking the callable.
    If processor returns False, callable will not be invoked.
    If processor returns a 2-tuple, it will use those as *args, **kwargs to invoke the method.
    """

    def __init__(self, event: Union[Callable, str], call: Callable = None, processor: Callable = None):
        self.event = event
        self.call = call
        self.processor = processor


class Action(RegistryObject, ABC):
    """
    Action can perform something on the system through "execute" method, or any other.
    """

    @property
    @abstractmethod
    def event_bindings(self) -> Union[Iterable[Union[Callable, str, EventBinding]], Union[Callable, str, EventBinding]]:
        """
        The event bindings that should trigger the action.

        A binding can be defined as a string matching the event name and will register the "execute" method on event
        bus.

        A binding can also be defined as EventBinding instance.
        """

    @property
    def order(self) -> int:
        """
        The order index related to this action.

        When many actions are registered on the same event name, they will be executed in sequence following the
        natural ordering from this value.
        """
        return 0

    @property
    def disabled(self) -> bool:
        """
        Check if the action is disabled.
        """
        return False


class InitializableAction(Action, ABC):  # pylint:disable=abstract-method
    """
    An action supporting an initialize singleton method.
    """

    def __init__(self):
        self.initialized = False

    def initialize(self):
        """
        Initialize method, invoked before first event binding execution.
        """

    def destroy(self):
        """
        destroy method, invoked on shutdown.
        """


class WatchSupport(ABC):
    """
    Watch support
    """

    @abstractmethod
    def start_watching(self):
        """
        Start watching
        """

    @abstractmethod
    def stop_watching(self):
        """
        Stop watching
        """

    @abstractmethod
    def join_watching(self):
        """
        Join watching
        """


class AbstractTemplateAction(InitializableAction, ABC):  # pylint:disable=abstract-method
    """
    Abstract action to render templates based on filename suffixes, supporting incoming events from file feature.
    """

    def __init__(self):
        super().__init__()
        self.template_finder = None  # type: TemplateFinder
        register_project_cache(self._cache_key)

    @property
    def _cache_key(self):
        """
        This cache is used to store the rendered data of all targets.
        When a template is deleted, target file will remove only if content is still the same than the generated one,
        thanks to this cache.
        """
        return "template.target." + self.__class__.__name__

    @property
    def event_bindings(self):
        def file_found_processor(file: str):
            template = file
            target = self.template_finder.get_target(template)
            if target:
                return (), {"template": template, "target": target}
            return None

        def file_delete_processor(file: str):
            template = file
            target = self.template_finder.get_target(template)
            if target and not self._target_is_modified(template, target):
                return (), {"template": template, "target": target}
            return None

        def file_generated_processor(source: str, target: str):
            template = target
            target = self.template_finder.get_target(template)
            if target:
                return (), {"template": template, "target": target}
            return None

        return (EventBinding(events.file.found, processor=file_found_processor),
                EventBinding(events.file.deleted, call=self.delete, processor=file_delete_processor),
                EventBinding("file:generated", processor=file_generated_processor))

    def initialize(self):
        self.template_finder = self._build_template_finder()

    def delete(self, template: str, target: str):
        """
        Delete a rendered template
        """
        if os.path.exists(target):
            force_remove(target)
            context.log.warning("%s removed", target)
            caches.get(self._cache_key).pop(target)
            events.file.deleted(target)

    def execute(self, template: str, target: str, migrate_retries_count=0, original_template=None):
        """
        Render a template
        """
        if os.path.abspath(os.path.join(os.getcwd(), target)) in config.files:
            # Never write a file that match ddb configuration file
            return

        if original_template is None:
            original_template = template

        try:
            for rendered, destination in self._render_template(template, target):
                if original_template != template and 'autofix' in config.args and config.args.autofix:
                    context.logger.info("[autofix]: %s", original_template)
                    copy_if_different(template, original_template, log=True)
                written = False
                if not isinstance(rendered, bool):
                    is_bynary = isinstance(rendered, (bytes, bytearray))
                    written = write_if_different(destination, rendered,
                                                 'rb' if is_bynary else 'r',
                                                 'wb' if is_bynary else 'w',
                                                 log_source=original_template)
                    caches.get(self._cache_key).set(target, rendered)
                context.mark_as_processed(template, destination)

                if written or rendered is True or config.eject:
                    events.file.generated(source=original_template, target=destination)
        except Exception as render_error:  # pylint:disable=broad-except
            if migrate_retries_count > 50:
                raise render_error
            try:
                template = self._autofix_render_error(template, target, original_template, render_error)
            except Exception as migrate_error:
                context.log.warning("Automatic template migration has failed for \"%s\": %s", template, migrate_error)
                raise render_error from migrate_error
            if not template:
                raise render_error

            migrate_retries_count += 1
            self.execute(template, target, migrate_retries_count, original_template)

    def _target_is_modified(self, template: str, target: str) -> bool:
        rendered = caches.get(self._cache_key).get(target)
        if rendered is None:
            return True
        if not os.path.exists(target):
            caches.get(self._cache_key).pop(target)
            return True
        is_bynary = isinstance(rendered, (bytes, bytearray))
        read_encoding = None if is_bynary else "utf-8"
        with open(target, mode='rb' if is_bynary else 'r', encoding=read_encoding) as read_file:
            existing_data = read_file.read()
        return existing_data != rendered

    @abstractmethod
    def _build_template_finder(self) -> TemplateFinder:
        """
        Build the template finder.
        """

    @abstractmethod
    def _render_template(self, template: str, target: str) -> Iterable[Tuple[Union[str, bytes, bool], str]]:
        """
        Perform template rendering to a string.
        """

    def _autofix_render_error(self,  # pylint:disable=no-self-use
                              template: str,
                              target: str,
                              original_template: str,
                              render_error: Exception) -> Optional[str]:
        """
        Try to automatically migrate this template in a temporary file when an error occurs.
        :return a temporary filepath to render again, if template has been fixed.
        """
        return None



# -*- coding: utf-8 -*-
from marshmallow import fields

from ddb.feature.schema import FeatureSchema


class JinjaSchema(FeatureSchema):
    """
    Jinja schema.
    """
    suffixes = fields.List(fields.String(), default=[".jinja"])
    extensions = fields.List(fields.String(), default=[".*", ""])
    includes = fields.List(fields.String())  # default is build automatically from suffixes value
    excludes = fields.List(fields.String())
    options = fields.Dict(fields.String(required=True), fields.Field(required=True), allow_none=True, default=None)



from ddb.__main__ import register_default_caches, clear_caches
from ddb.command import LifecycleCommand
from ddb.config import config
from ddb.event import bus
from ddb.phase import DefaultPhase, phases


def test_lifecycle():
    register_default_caches()

    phases.register(DefaultPhase("step1"))
    phases.register(DefaultPhase("step2"))

    config.args.clear_cache = False

    events = []

    bus.on(None, lambda event: events.append(event))

    command = LifecycleCommand("test", "TestCommand", "step1", "step2", DefaultPhase("step3"))

    command.execute()

    assert events == ["phase:step1", "phase:step2", "phase:step3"]
    events = []

    command.execute()
    assert events == ["phase:step1", "phase:step2", "phase:step3"]


def test_lifecycle_run_once():
    register_default_caches()
    clear_caches()

    config.args.clear_cache = False

    phases.register(DefaultPhase("step1", run_once=True))
    phases.register(DefaultPhase("step2"))

    events = []

    bus.on(None, lambda event, *args, **kwargs: events.append(event))

    command = LifecycleCommand("test", "TestCommand", "step1", "step2", "step1", DefaultPhase("step3"))

    command.execute()

    assert events == ["phase:step1", "phase:step2", "phase:step3"]
    events = []

    command.execute()
    assert events == ["phase:step2", "phase:step3"]



import glob


def found_with_glob(value):
    """
    Check if at least one file match the given glob expression.
    :param value: glob expression
    """
    for _ in glob.glob(value):
        return True
    return False



# -*- coding: utf-8 -*-

from ddb.event import Bus


def test_with_event_name():
    bus = Bus()

    custom_event_listeners = []

    def listener(data):
        custom_event_listeners.append(data)

    bus.on("custom:event", listener)
    bus.emit("custom:event", "testing")
    bus.emit("custom:event2", "another")

    bus.off("custom:event", listener)
    bus.emit("custom:event", "more")

    assert custom_event_listeners == ["testing"]


def test_without_event_name():
    bus = Bus()

    custom_event_listeners = []

    def listener(event, data):
        custom_event_listeners.append((event, data))

    bus.on(None, listener)
    bus.emit("custom:event", "testing")
    bus.emit("custom:event2", "another")

    bus.off(None, listener)
    bus.emit("custom:event", "more")

    assert custom_event_listeners == [("custom:event", "testing"), ("custom:event2", "another")]



import os
import pytest

from ddb.__main__ import main, reset
from ddb.config import config, Config
from tests.utilstest import setup_cfssl


class TestTraefik:
    @pytest.mark.docker
    def test_traefik_extra_services_with_remove(self, project_loader, module_scoped_container_getter):
        project_loader("extra-services")

        setup_cfssl(module_scoped_container_getter)

        main(["configure"])

        api_toml = os.path.join(config.paths.home, "traefik", "config", "sub.project.test.extra-service.api.toml")
        cert = os.path.join(config.paths.home, "certs", "sub.project.test.crt")
        key = os.path.join(config.paths.home, "certs", "sub.project.test.key")
        web_toml = os.path.join(config.paths.home, "traefik", "config", "rule.project.test.extra-service.web.toml")

        assert os.path.exists(api_toml)
        assert os.path.exists(cert)
        assert os.path.exists(key)
        assert os.path.exists(web_toml)

        with open("ddb.yml", "w"):
            pass

        Config.defaults = {'defaults': {'fail_fast': True}}

        setup_cfssl(module_scoped_container_getter)

        main(["configure"])

        assert not os.path.exists(api_toml)
        assert not os.path.exists(cert)
        assert not os.path.exists(key)
        assert not os.path.exists(web_toml)

    @pytest.mark.docker
    def test_traefik_extra_services_without_remove(self, project_loader, module_scoped_container_getter):
        project_loader("extra-services")

        setup_cfssl(module_scoped_container_getter)

        main(["configure"])

        api_toml = os.path.join(config.paths.home, "traefik", "config", "sub.project.test.extra-service.api.toml")
        cert = os.path.join(config.paths.home, "certs", "sub.project.test.crt")
        key = os.path.join(config.paths.home, "certs", "sub.project.test.key")
        web_toml = os.path.join(config.paths.home, "traefik", "config", "rule.project.test.extra-service.web.toml")

        assert os.path.exists(api_toml)
        assert os.path.exists(cert)
        assert os.path.exists(key)
        assert os.path.exists(web_toml)

        main(["configure"])

        assert os.path.exists(api_toml)
        assert os.path.exists(cert)
        assert os.path.exists(key)
        assert os.path.exists(web_toml)



# -*- coding: utf-8 -*-
from marshmallow import fields

from ddb.feature.schema import FeatureSchema


class GitignoreSchema(FeatureSchema):
    """
    Git schema.
    """
    disabled = fields.Bool(default=None)  # default is set in feature _configure_defaults
    enforce = fields.List(fields.String(), default=["*ddb.local.*"])
    markers = fields.List(fields.String(), default=["inetum-orleans/docker-devbox", "gfi-centre-ouest/docker-devbox"])



# -*- coding: utf-8 -*-
from marshmallow import fields, Schema

from ddb.feature.schema import FeatureSchema


class CopySpecSchema(Schema):
    """
    Copy to build context schema.
    """
    source = fields.String(required=True)
    destination = fields.String(required=True, default=".")
    filename = fields.String(required=False)
    dispatch = fields.List(fields.String(), required=False)


class CopySchema(FeatureSchema):
    """
    Copy schema.
    """
    specs = fields.List(fields.Nested(CopySpecSchema(), default=None))



import os

import yaml

from ddb.__main__ import main
from tests.utilstest import expect_gitignore


class TestFilesGenerated:
    def test_ensure_chaining(self, project_loader):
        project_loader("ensure-chaining")

        main(["configure"])

        main(["activate"])

        assert os.path.exists("test.dev.yml.jsonnet")
        assert not os.path.exists("test.dev.yml.jinja")
        with open("test.dev.yml.jsonnet", "r") as dockerfile:
            data = dockerfile.read()
            assert data == '{["e" + "n" + "v"]: "dev"}\n'

        assert os.path.exists("test.dev.yml")
        with open("test.dev.yml", "r") as dockerfile:
            data = dockerfile.read()
            assert yaml.load(data, yaml.SafeLoader) == {"env": "dev"}

        assert os.path.exists("test.yml")
        with open("test.yml", "r") as dockerfile:
            data = dockerfile.read()
            assert yaml.load(data, yaml.SafeLoader) == {"env": "dev"}

        assert os.path.islink("test.yml")

        assert os.path.exists(".gitignore")
        assert expect_gitignore('.gitignore', '/test.dev.yml.jsonnet', '/test.dev.yml', '/test.yml')

    def test_ensure_chaining_with_custom_dependencies(self, project_loader):
        project_loader("ensure-chaining-with-custom-dependencies")

        main(["configure"])

        main(["activate"])

        assert not os.path.exists("test.dev.yml.jsonnet")
        assert os.path.exists("test.dev.yml.jinja")
        with open("test.dev.yml.jinja", "r") as dockerfile:
            data = dockerfile.read().splitlines()
            assert data == ['{', '   "env": "{{ core.env.current }}"', '}']

        assert os.path.exists("test.dev.yml")
        with open("test.dev.yml", "r") as dockerfile:
            data = dockerfile.read()
            assert yaml.load(data, yaml.SafeLoader) == {"env": "dev"}

        assert os.path.exists("test.yml")
        with open("test.yml", "r") as dockerfile:
            data = dockerfile.read()
            assert yaml.load(data, yaml.SafeLoader) == {"env": "dev"}

        assert os.path.islink("test.yml")

        assert os.path.exists(".gitignore")
        assert expect_gitignore('.gitignore', '/test.dev.yml.jinja', '/test.dev.yml', '/test.yml')

    def test_directories_overriden(self, project_loader):
        project_loader("directories-overriden")

        assert os.path.isdir("test")

        main(["configure"])

        main(["activate"])

        assert not os.path.isdir("test")
        assert os.path.isfile("test")

    def test_remove_target_on_source_removed(self, project_loader):
        project_loader("remove_target_on_source_removed")

        main(["configure"])

        main(["activate"])

        assert os.path.islink("test")

        os.remove("test.dev")
        os.remove("test2.jinja")

        main(["configure"])

        assert not os.path.exists("test")
        assert not os.path.exists("test2")

    def test_do_not_remove_modified_target_on_source_removed_symlink(self, project_loader):
        project_loader("remove_target_on_source_removed")

        main(["configure"])

        assert os.path.islink("test")
        assert os.path.isfile("test2")

        os.remove("test")
        os.remove("test.dev")
        os.remove("test2.jinja")

        with open("test", "w") as fp:
            fp.write("test")

        main(["configure"])

        assert os.path.exists("test")
        assert not os.path.exists("test2")



# -*- coding: utf-8 -*-
import os

from marshmallow import fields, Schema

from ddb.feature.schema import FeatureSchema
from ...config import config


def _get_path_default():
    _path_default = []

    if config.paths.project_home:
        _path_default.append(".bin")
        _path_default.append("bin")

    if config.paths.home:
        _path_default.append(os.path.join(config.paths.home, ".bin"))

    if config.paths.ddb_home:
        _path_default.append(os.path.join(config.paths.ddb_home, ".bin"))

    return _path_default


class PathSchema(Schema):
    """
    Shell schema.
    """
    directories = fields.List(fields.String(), required=False, default=_get_path_default)
    prepend = fields.Boolean(default=True)


class ShellSchema(FeatureSchema):
    """
    Shell schema.
    """
    shell = fields.String(required=True)
    path = fields.Nested(PathSchema, required=True, default=PathSchema())
    envignore = fields.List(fields.String(), required=False, default=[
        "PYENV_*", "_", "PS1", "PS2", "PS3", "PS4", "PWD"
    ])
    aliases = fields.Dict(required=False, default={})
    global_aliases = fields.List(fields.String(), required=False, default=[])



# -*- coding: utf-8 -*-
import os
import re
from pathlib import Path
from typing import Union, Iterable, Tuple, Optional, Set

from jinja2 import Environment, FileSystemLoader, StrictUndefined

from ddb.config import config, migrations
from ddb.utils.file import TemplateFinder, SingleTemporaryFile, get_single_temporary_file_directory
from . import filters, tests
from ...action.action import AbstractTemplateAction
from ...config.migrations import AbstractPropertyMigration

custom_filters = vars(filters)
for k in tuple(custom_filters.keys()):
    if k.startswith("__"):
        del custom_filters[k]

custom_tests = vars(tests)
for k in tuple(custom_tests.keys()):
    if k.startswith("__"):
        del custom_tests[k]


class JinjaAction(AbstractTemplateAction):
    """
    Render jinja templates based on filename suffixes.
    """

    def __init__(self):
        super().__init__()
        self.env = None  # type: Environment
        self.context = None  # type: dict
        self._rootpath = None  # type: str
        self._migrationpath = None  # type: str

    @property
    def name(self) -> str:
        return "jinja:render"

    def _build_template_finder(self):
        return TemplateFinder(config.data.get("jinja.includes"),
                              config.data.get("jinja.excludes"),
                              config.data.get("jinja.suffixes"))

    def initialize(self):
        super().initialize()

        self._rootpath = self.template_finder.rootpath
        self._migrationpath = get_single_temporary_file_directory("ddb", "migration", "jinja")

        effective_options = {
            "loader": FileSystemLoader([self._rootpath, self._migrationpath]),
            "undefined": StrictUndefined,
            "keep_trailing_newline": True
        }

        options = config.data.get('jinja.options')
        if options:
            effective_options.update(options)

        self.env = Environment(**effective_options)

        self.env.filters.update(custom_filters)
        self.env.tests.update(custom_tests)

        self.context = config.data.copy()
        self.context['_config'] = {}
        self.context['_config.eject'] = config.eject
        self.context['_config.args'] = vars(config.args)
        self.context['_config.unknown_args'] = config.unknown_args

    def _render_template(self, template: str, target: str) -> Iterable[Tuple[Union[str, bytes, bool], str]]:
        if template.startswith(self._migrationpath):
            template_name = os.path.relpath(os.path.normpath(template),
                                            os.path.normpath(str(self._migrationpath)))

        else:
            template_name = os.path.relpath(os.path.normpath(template),
                                            os.path.normpath(str(self._rootpath)))

        template_name = Path(template_name).as_posix()
        jinja = self.env.get_template(template_name)
        yield jinja.render(**self.context), target

    def _autofix_render_error(self,
                              template: str,
                              target: str,
                              original_template: str,
                              render_error: Exception) -> Optional[str]:
        property_migration_set = self._find_property_migration_set(render_error)

        if property_migration_set:
            altered_template_data = self._apply_migration_set(template,
                                                              original_template,
                                                              property_migration_set)

            if altered_template_data:
                with SingleTemporaryFile("ddb", "migration", "jinja",
                                         mode="w",
                                         prefix="",
                                         suffix="." + os.path.basename(original_template),
                                         encoding="utf-8") as tmp_file:
                    tmp_file.write(altered_template_data)
                    return tmp_file.name

        return None

    @staticmethod
    def _find_property_migration_set(render_error: Exception) -> Set[AbstractPropertyMigration]:
        property_migration_set = set()

        undefined_match = re.match("'(.*)' is undefined", str(render_error))
        if undefined_match:
            property_name = undefined_match.group(1)
            property_migration = migrations.get_migration_from_old_config_key(property_name)
            if property_migration:
                property_migration_set.add(property_migration)
        else:
            dict_object_match = re.match("'dict object' has no attribute '(.*)'", str(render_error))
            if dict_object_match:
                property_contains = f".{dict_object_match.group(1)}"

                for property_migration in migrations.get_history():
                    if isinstance(property_migration, AbstractPropertyMigration) \
                            and not property_migration.requires_value_migration \
                            and property_contains in property_migration.old_config_key:
                        property_migration_set.add(property_migration)

        return property_migration_set

    def _apply_migration_set(self,
                             template: str,
                             original_template: str,
                             property_migration_set: Iterable[AbstractPropertyMigration]):
        with open(template, "r", encoding="utf-8") as template_file:
            altered = False
            altered_template_data = ""
            variable_content = ""
            in_variable = False
            previous_lex_lineno = None
            previous_lex_value = None

            for lex_lineno, lex_type, lex_value in self.env.lex(self.env.preprocess(template_file.read())):
                lex_value = self._fix_lex_value(lex_value, lex_lineno, previous_lex_value, previous_lex_lineno)

                if lex_type in ('variable_end', 'block_end'):
                    altered_variable_content = JinjaAction._replace_variable_content(variable_content,
                                                                                     property_migration_set,
                                                                                     original_template)
                    if altered_variable_content != variable_content:
                        altered = True

                    altered_template_data = altered_template_data + altered_variable_content
                    variable_content = ""
                    in_variable = False

                if in_variable:
                    variable_content = variable_content + lex_value
                else:
                    altered_template_data = altered_template_data + lex_value

                if lex_type in ('variable_begin', 'block_begin'):
                    in_variable = True
                    variable_content = ""

                previous_lex_lineno = lex_lineno
                previous_lex_value = lex_value

        altered_template_data = altered_template_data + variable_content
        return altered_template_data if altered else None

    @staticmethod
    def _replace_variable_content(variable_content, property_migration_set, original_template):
        for property_migration in property_migration_set:
            if property_migration and not property_migration.requires_value_migration:
                new_variable_content = variable_content \
                    .replace(property_migration.old_config_key, property_migration.new_config_key)
                if new_variable_content != variable_content:
                    variable_content = new_variable_content
                    property_migration.warn(original_template)
        return variable_content

    @staticmethod
    def _fix_lex_value(lex_value, lex_lineno, previous_lex_value, previous_lex_lineno):
        if previous_lex_lineno is not None and lex_lineno > previous_lex_lineno:
            previous_eol_count = previous_lex_value.count('\n')
            lex_value = (lex_lineno - previous_lex_lineno - previous_eol_count) * '\n' + lex_value
        return lex_value



# -*- coding: utf-8 -*-
import logging
import os

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from watchdog.utils.delayed_queue import DelayedQueue

from ddb.action import InitializableAction
from ddb.action.action import WatchSupport
from ddb.cache import caches, register_project_cache
from ddb.config import config
from ddb.context import context
from ddb.event import events
from ddb.utils.file import FileWalker


class FileWalkAction(InitializableAction, WatchSupport):
    """
    Emit events for files inside project matching patterns.
    """

    def __init__(self):
        super().__init__()
        self.queue = None  # type: DelayedQueue
        self.file_walker = None  # type: FileWalker
        self.observer = None  # type: Observer

    @property
    def name(self) -> str:
        return "file:walk"

    @property
    def event_bindings(self):
        return events.phase.configure

    def initialize(self):
        self.file_walker = FileWalker(config.data.get("file.includes"),
                                      config.data.get("file.excludes"),
                                      config.data.get("file.suffixes"))

        register_project_cache("file")

    def destroy(self):
        if caches.has("file"):
            caches.unregister("file", callback=lambda c: c.close())

    def execute(self):
        """
        Execute action
        """
        cache = caches.get("file")
        found_files = set()

        for file in self.file_walker.items:
            cache.set(file, None)
            found_files.add(file)

        for cached_file in cache.keys():
            if cached_file not in found_files:
                cache.pop(cached_file)
                events.file.deleted(cached_file)

        for file in self.file_walker.items:
            events.file.found(file)

        cache.flush()

    def start_watching(self):
        self.observer = Observer()
        self.observer.schedule(ObserverHandler(self), str(self.file_walker.rootpath), self.file_walker.recursive)
        self.observer.start()

    def stop_watching(self):
        self.observer.stop()

    def join_watching(self):
        self.observer.join()


class ObserverHandler(FileSystemEventHandler):
    """
    File watcher handler
    """

    def __init__(self, action: FileWalkAction):
        self.action = action

    def emit_if_not_filtered(self, func, file):
        """
        Emit given event if file is not filtered by file_walker
        """
        if not file.endswith('~') and not self.action.file_walker.is_source_filtered(file):
            logging.getLogger("ddb.watch").debug("%s", file)
            context.mark_as_unprocessed(file)
            func(file=file)

    def on_modified(self, event):
        if not event.is_directory:
            source = os.path.relpath(event.src_path)
            self.emit_if_not_filtered(events.file.found, source)

    def on_created(self, event):
        if not event.is_directory:
            source = os.path.relpath(event.src_path)
            self.emit_if_not_filtered(events.file.found, source)

    def on_deleted(self, event):
        if not event.is_directory:
            source = os.path.relpath(event.src_path)
            self.emit_if_not_filtered(events.file.deleted, source)

    def on_moved(self, event):
        if not event.is_directory:
            source = os.path.relpath(event.src_path)
            destination = os.path.relpath(event.dest_path)
            self.emit_if_not_filtered(events.file.found, destination)
            self.emit_if_not_filtered(events.file.deleted, source)
